import colorsys
from ._colour_schemes import hexadecimal_to_name, name_to_hexadecimal, colour_schemes
from ._colourize import colourize

DEFAULT_INCREASE_RATIO = 0.2
DEFAULT_INCREASE_AMOUNT = 0.1


def scale(x, minimum, maximum):
	return (x - minimum) / (maximum - minimum)


def limit(x, minimum=0.0, maximum=1.0):
	return min(maximum, max(minimum, x))


class Colour:
	def __init__(
			self, obj=None, red=None, green=None, blue=None, hexadecimal=None,
			hue=None, saturation=None, lightness=None, value=None,
			min_value=0.0, max_value=1.0, name=None, id=None, scheme=None,
			weight=1.0
	):
		"""
		:type obj: str or NoneType or Colour
		:type red: float or NoneType
		:type green: float or NoneType
		:type blue: float or NoneType
		:type min_value: float
		:type max_value: float
		:type name: NoneType or str
		:type id: NoneType or int
		:type scheme: NoneType or .Scheme.Scheme
		"""
		self._id = id
		self._scheme = scheme
		self._weight = weight

		if obj is not None:
			if isinstance(obj, self.__class__):
				red, green, blue = obj.rgb

			elif isinstance(obj, str):
				if obj.startswith('#'):
					hexadecimal = obj

				else:
					name = obj

			elif isinstance(obj, (list, tuple)):
				red, green, blue = obj

			else:
				raise TypeError(f'obj should be one of str, Colour, tuple, or list but it is {type(obj)}')
			min_value = 0.0
			max_value = 1.0

		if red is not None and green is not None and blue is not None:
			pass

		elif hue is not None and saturation is not None and lightness is not None:
			red, green, blue = colorsys.hls_to_rgb(h=hue, l=lightness, s=saturation)
			min_value = 0.0
			max_value = 1.0

		elif hue is not None and saturation is not None and value is not None:
			red, green, blue = colorsys.hsv_to_rgb(h=hue, s=saturation, v=value)
			min_value = 0.0
			max_value = 1.0

		elif hexadecimal is not None:
			red, green, blue = self.convert_hexadecimal_to_rgb(hexadecimal=hexadecimal)
			min_value = 0.0
			max_value = 255.0

		elif name is not None:
			hexadecimal = name_to_hexadecimal[name.lower()]
			red, green, blue = self.convert_hexadecimal_to_rgb(hexadecimal=hexadecimal)
			min_value = 0.0
			max_value = 255.0

		else:
			raise ValueError(f'obj: "{obj}" not acceptable. It is of type "{type(obj)}"')

		self._red = scale(x=red, minimum=min_value, maximum=max_value)
		self._green = scale(x=green, minimum=min_value, maximum=max_value)
		self._blue = scale(x=blue, minimum=min_value, maximum=max_value)
		self._name = name

	@property
	def scheme(self):
		"""
		:rtype: .Scheme.Scheme or NoneType
		"""
		return self._scheme

	@property
	def id(self):
		"""
		:rtype: int or NoneTy[e
		"""
		return self._id

	def use(self, log=None):
		"""
		logs the usage of a colour
		:type log:
		:rtype: NoneType
		"""
		if self.scheme is not None:
			self.scheme.use(colour=self, log=log)

	@property
	def logs(self):
		if self.scheme is not None:
			return self.scheme._usage_logs[self.id]
		else:
			return None

	@property
	def usage(self):
		"""
		:rtype: int
		"""
		if self.scheme is not None:
			return self.scheme.get_usage(colour=self)
		else:
			return 0

	def __delete_identity(self):
		self._name = None
		self._id = None
		self._scheme = None

	@property
	def red(self):
		return limit(x=self._red)

	@red.setter
	def red(self, red):
		self.__delete_identity()
		self._red = red

	@property
	def green(self):
		return limit(x=self._green)

	@green.setter
	def green(self, green):
		self.__delete_identity()
		self._green = green

	@property
	def blue(self):
		return limit(x=self._blue)

	@blue.setter
	def blue(self, blue):
		self.__delete_identity()
		self._blue = blue

	def copy(self, keep_id=True):
		if keep_id:
			return self.__class__(
				red=self.red, green=self.green, blue=self.blue, name=self._name, id=self._id, scheme=self._scheme,
				weight=self._weight
			)
		else:
			return self.__class__(
				red=self.red, green=self.green, blue=self.blue, name=self._name, id=None, scheme=None,
				weight=self._weight
			)

	def __str__(self):
		return f'{self.name}: {self.hexadecimal}'

	def __repr__(self):
		return str(self)

	@staticmethod
	def _get_names():
		"""
		:rtype: dict[str,str]
		"""
		return hexadecimal_to_name

	@staticmethod
	def _get_hexadecimals():
		"""
		:rtype: dict[str, str]
		"""
		return name_to_hexadecimal

	@classmethod
	def get_standard_colours(cls):
		"""
		:rtype: list[Colour]
		"""
		return [cls(hexadecimal=hexadecimal, name=name) for hexadecimal, name in cls._get_names().items()]

	@staticmethod
	def get_schemes():
		"""
		:rtype: dict[str, list[str]]
		"""
		return colour_schemes.copy()

	@staticmethod
	def convert_hexadecimal_to_rgb(hexadecimal):
		if hexadecimal.startswith('#'):
			hexadecimal = hexadecimal[1:]

		if len(hexadecimal) == 3:
			hexadecimal = \
				hexadecimal[0] + hexadecimal[0] + \
				hexadecimal[1] + hexadecimal[1] + \
				hexadecimal[2] + hexadecimal[2]

		if len(hexadecimal) != 6:
			raise ValueError('The hex value should be 3 or 6 characters long!')

		red = int(f"0x{hexadecimal[0:2]}", 16)
		green = int(f"0x{hexadecimal[2:4]}", 16)
		blue = int(f"0x{hexadecimal[4:6]}", 16)
		return red, green, blue

	@property
	def name(self):
		if self._name is None:
			self._name = self.find_nearest(colours=self.get_standard_colours()).name
		return self._name

	@property
	def rgb(self):
		"""
		:rtype: tuple
		"""
		return self.red, self.green, self.blue

	@property
	def hsl(self):
		"""
		:rtype: tuple
		"""
		h, l, s = colorsys.rgb_to_hls(r=self.red, g=self.green, b=self.blue)
		return h, s, l

	@property
	def yig(self):
		"""
		:rtype: tuple
		"""
		return colorsys.rgb_to_yiq(r=self.red, g=self.green, b=self.blue)

	@property
	def hsv(self):
		"""
		:rtype: tuple
		"""
		return colorsys.rgb_to_hsv(r=self.red, g=self.green, b=self.blue)

	@property
	def hue(self):
		"""
		:rtype: float
		"""
		return self.hsl[0]

	@hue.setter
	def hue(self, hue):
		_, saturation, lightness = self.hsl
		hue = hue % 1.0
		self.__delete_identity()
		self.red, self._green, self._blue = colorsys.hls_to_rgb(h=hue, l=lightness, s=saturation)

	@property
	def saturation(self):
		"""
		:rtype: float
		"""
		return self.hsl[1]

	@saturation.setter
	def saturation(self, saturation):
		saturation = min(1.0, max(0, saturation))
		hue, _, lightness = self.hsl
		self._name = None
		self.red, self._green, self._blue = colorsys.hls_to_rgb(h=hue, l=lightness, s=saturation)

	def set_lightness_and_saturation(self, lightness, saturation):
		hue = self.hue
		self.red, self._green, self._blue = colorsys.hls_to_rgb(h=hue, l=lightness, s=saturation)

	@property
	def lightness(self):
		"""
		:rtype: float
		"""
		return self.hsl[2]

	@lightness.setter
	def lightness(self, lightness):
		lightness = min(1.0, max(0.0, lightness))
		hue, saturation, _ = self.hsl
		self._name = None
		self.red, self._green, self._blue = colorsys.hls_to_rgb(h=hue, l=lightness, s=saturation)

	@property
	def value(self):
		"""
		:rtype: float
		"""
		return self.hsv[2]

	@value.setter
	def value(self, value):
		value = min(1.0, max(0.0, value))
		hue, saturation, _ = self.hsv
		self._name = None
		self.red, self._green, self._blue = colorsys.hsv_to_rgb(h=hue, v=value, s=saturation)

	@property
	def hexadecimal(self):
		"""
		:rtype: str
		"""
		r = int(min(max(0.0, self.red * 255), 255))
		g = int(min(max(0.0, self.green * 255), 255))
		b = int(min(max(0.0, self.blue * 255), 255))
		return '#{:02x}{:02x}{:02x}'.format(r, g, b)

	def get_hexadecimal(self, opacity=None):
		"""
		:rtype: str
		"""
		if opacity is None:
			result = self.hexadecimal
		else:
			o = int(min(max(0.0, opacity * 255), 255))
			result = '{}{:02x}'.format(self.hexadecimal, o)
		return result

	def get_distance(self, other):
		"""
		:type other: Colour
		:rtype: float
		"""
		red2 = (self.red - other.red) ** 2
		green2 = (self.green - other.green) ** 2
		blue2 = (self.blue - other.blue) ** 2
		return (red2 + green2 + blue2) ** 0.5

	def find_nearest(self, colours):
		"""
		:type colours: list[Colour]
		:rtype:
		"""
		return sorted(colours, key=lambda x: self.get_distance(other=x))[0]

	def limit(self):
		self._red = self.red
		self._green = self.green
		self._blue = self.blue

	@classmethod
	def _tuple_as_colour(cls, other):
		if isinstance(other, (tuple, list)):
			if len(other) == 3:
				other = cls(red=other[0], green=other[1], blue=other[2])
		return other

	def __add__(self, other):
		"""
		:type other: Colour
		:rtype: Colour
		"""
		other = self._tuple_as_colour(other=other)

		if isinstance(other, Colour):
			return self.__class__(
				red=self._red + other._red,
				green=self._green + other._green,
				blue=self._blue + other._blue,
				weight=self._weight
			)
		else:
			return self.__class__(
				red=self._red + other, green=self._green + other, blue=self._blue + other, weight=self._weight
			)

	def __sub__(self, other):
		"""
		:type other: Colour
		:rtype: Colour
		"""
		other = self._tuple_as_colour(other=other)

		return self.__class__(
			red=self._red - other._red,
			green=self._green - other._green,
			blue=self._blue - other._blue
		)

	def __mul__(self, other):
		"""
		:type other: float or tuple or Colour
		:rtype: Colour
		"""
		if isinstance(other, (list, tuple)):
			other = self._tuple_as_colour(other=other)

		if isinstance(other, Colour):
			return self.__class__(
				red=self.red * other.red,
				green=self.green * other.green,
				blue=self.blue * other.blue
			)
		else:
			return self.__class__(red=self._red, green=self._green, blue=self._blue, weight=self._weight * other)

	def __neg__(self):
		"""
		:rtype: Colour
		"""

		return self.__class__(red=1 - self.red, green=1 - self.green, blue=1 - self.blue)

	def __invert__(self):
		return -self

	def __and__(self, other):
		if isinstance(other, (int, float)):
			if other > 1 or other < 0:
				raise ValueError('other cannot be negative or greated than one.')
		other = self._tuple_as_colour(other=other)

		return self * other

	def __or__(self, other):
		if isinstance(other, (int, float)):
			if other > 1 or other < 0:
				raise ValueError('other cannot be negative or greated than one.')
		other = self._tuple_as_colour(other=other)

		return self + other

	def __hashkey__(self):
		return self.__class__.__name__, self.__getstate__()

	def __getstate__(self):
		return self.red, self.green, self.blue, self._name, self._id, self._weight, None

	def __setstate__(self, state):
		self._red, self._green, self._blue, self._name, self._id, self._weight, self._scheme = state

	@classmethod
	def _from_state(cls, state):
		red, green, blue, name, id, weight, _ = state
		return cls(red=red, green=green, blue=blue, id=id, weight=weight)

	def colourize(self, string, background=None):
		if background == 'auto':
			background = self.farthest_gray

		if background is None:
			bg_red, bg_green, bg_blue = None, None, None
		else:
			if not isinstance(background, self.__class__):
				background = self.__class__(obj=background)
			bg_red, bg_green, bg_blue = background.rgb

		return colourize(
			string=string, red=self.red, green=self.green, blue=self.blue,
			bg_red=bg_red, bg_green=bg_green, bg_blue=bg_blue
		)

	def colourize_background(self, string, text_colour='auto'):
		if isinstance(text_colour, str) and text_colour == 'auto':
			text_colour = self.farthest_gray

		if text_colour is not None:
			bg_red, bg_green, bg_blue = text_colour.rgb
		else:
			bg_red, bg_green, bg_blue = self.__class__(hexadecimal='#808080').rgb

		return colourize(
			string=string, bg_red=self.red, bg_green=self.green, bg_blue=self.blue,
			red=bg_red, green=bg_green, blue=bg_blue
		)

	def print(self, string, secondary='auto', end=None, main_colour='background'):
		if main_colour == 'background':
			print(self.colourize_background(string=string, text_colour=secondary), end=end)
		else:
			print(self.colourize(string=string, background=secondary), end=end)

	def display(self, string=None, secondary='auto', end=None, main_colour='background', length=None):
		string = string or f' {self.name} '
		if length is not None:
			string = ('{:^' + str(int(length)) + '}').format(string)
		self.print(string=string, secondary=secondary, end=end, main_colour=main_colour)

	def mix_with_gray(self, gray_weight=None):
		"""
		:type gray_weight: NoneType or float
		:rtype: Colour
		"""
		gray = self.nearest_gray
		if gray_weight:
			gray._weight = gray_weight
		return self.mix(colours=gray)

	def darken(self, ratio=DEFAULT_INCREASE_RATIO, amount=None, keep_id=True):
		"""
		:type ratio: float or NoneType
		:type amount: float or NoneType
		:type keep_id: bool
		:rtype: Colour
		"""
		ratio = min(1.0, max(-1.0, ratio))
		darker = self.copy(keep_id=keep_id)

		if amount is None:
			amount = self.lightness ** 0.5 * ratio

		darker.lightness = darker.lightness - amount
		return darker

	def lighten(self, ratio=DEFAULT_INCREASE_RATIO, amount=None, keep_id=True):
		"""
		:type ratio: float or NoneType
		:type amount: float or NoneType
		:type keep_id: bool
		:rtype: Colour
		"""
		ratio = min(1.0, max(-1.0, ratio))
		lighter = self.copy(keep_id=keep_id)

		if amount is None:
			amount = (1 - self.lightness) * ratio

		lighter.lightness = lighter.lightness + amount
		return lighter

	def saturate(self, ratio=DEFAULT_INCREASE_RATIO, amount=None, keep_id=True):
		"""
		:type ratio: float or NoneType
		:type amount: float or NoneType
		:type keep_id: bool
		:rtype: Colour
		"""
		ratio = min(1.0, max(-1.0, ratio))
		more_saturated = self.copy(keep_id=keep_id)

		if amount is None:
			amount = (1 - self.saturation) * ratio

		more_saturated.saturation = more_saturated.saturation + amount
		return more_saturated

	def pale(self, ratio=DEFAULT_INCREASE_RATIO, amount=None, keep_id=True):
		"""
		:type ratio: float or NoneType
		:type amount: float or NoneType
		:type keep_id: bool
		:rtype: Colour
		"""
		ratio = min(1.0, max(-1.0, ratio))
		less_saturated = self.copy(keep_id=keep_id)

		if amount is None:
			amount = self.saturation * ratio

		less_saturated.saturation = less_saturated.saturation - amount
		return less_saturated

	def darken_or_lighten(self, ratio=DEFAULT_INCREASE_RATIO, amount=None, keep_id=True):
		"""
		:type ratio: float or NoneType
		:type amount: float or NoneType
		:type keep_id: bool
		:rtype: Colour
		"""
		if self.lightness <= 0.5:
			return self.lighten(ratio=ratio, amount=amount, keep_id=keep_id)
		else:
			return self.darken(ratio=ratio, amount=amount, keep_id=keep_id)

	def blacken_or_whiten(self):
		"""
		:rtype: Colour
		"""
		return self.farthest_gray

	brighten = lighten

	def increase_hue(self, amount=DEFAULT_INCREASE_AMOUNT, keep_id=False):
		"""
		:type amount: float
		:type keep_id: bool
		:type keep_id: bool
		:rtype: Colour
		"""
		changed = self.copy(keep_id=keep_id)
		changed.hue = changed.hue + amount
		return changed

	@property
	def farthest_gray(self):
		"""
		:rtype: Colour
		"""
		if self.lightness < 0.5:
			return self.__class__(hexadecimal='#FFFFFF', weight=self._weight)
		else:
			return self.__class__(hexadecimal='#000000', weight=self._weight)

	@property
	def nearest_gray(self, keep_id=False):
		"""
		:type keep_id: bool
		:rtype: Colour
		"""
		gray = self.copy(keep_id=keep_id)
		gray.saturation = 0
		return gray

	@property
	def nearest_red(self, keep_id=False):
		"""
		:type keep_id: bool
		:rtype: Colour
		"""
		red = self.copy(keep_id=keep_id)
		red.hue = 0
		return red

	@property
	def nearest_green(self, keep_id=False):
		"""
		:type keep_id: bool
		:rtype: Colour
		"""
		green = self.copy(keep_id=keep_id)
		green.hue = 1.0/3.0
		return green

	@property
	def nearest_blue(self, keep_id=False):
		"""
		:type keep_id: bool
		:rtype: Colour
		"""
		blue = self.copy(keep_id=keep_id)
		blue.hue = 2.0/3.0
		return blue

	def reverse_lightness(self, keep_id=True):
		"""
		:type keep_id: bool
		:rtype: Colour
		"""
		_reversed = self.copy(keep_id=keep_id)
		_reversed.lightness = 1 - _reversed.lightness
		return _reversed

	def invert(self):
		"""
		:rtype: Colour
		"""
		result = self.__class__(red=1 - self.red, green=1 - self.green, blue=1 - self.blue)
		return result

	def mix(self=None, colours=None):
		"""
		:type colours: NoneType or Colour or list[Colour]
		:rtype: Colour
		"""
		if colours is None or colours == []:
			return self.copy(keep_id=True)
		else:
			if isinstance(colours, Colour):
				colours = [colours]

		if self is not None:
			colours = [self] + colours

		red = green = blue = lightness = saturation = 0.0
		id = colours[0].id
		scheme = colours[0].scheme
		for colour in colours:
			red += colour.red * colour._weight
			green += colour.green * colour._weight
			blue += colour.blue * colour._weight
			#lightness += colour.lightness * colour._weight
			#saturation += colour.saturation * colour._weight
			if id != colour.id:
				id = None
				scheme = None

		total_weight = sum([colour._weight for colour in colours])
		red = red / total_weight
		green = green / total_weight
		blue = blue / total_weight
		#lightness = lightness / total_weight
		#saturation = saturation / total_weight
		if self is not None:
			result = self.__class__(
				red=red, green=green, blue=blue, min_value=0.0, max_value=1.0, id=id, scheme=scheme,
				weight=total_weight
			)
		else:
			result = Colour(
				red=red, green=green, blue=blue, min_value=0.0, max_value=1.0, id=id, scheme=scheme,
				weight=total_weight
			)

		#result.set_lightness_and_saturation(lightness=lightness, saturation=saturation)
		return result

	@property
	def weight(self):
		"""
		:rtype: float
		"""
		return self._weight

	@weight.setter
	def weight(self, weight):
		"""
		:type weight: float
		"""
		self._weight = weight

	def __eq__(self, other):
		if not isinstance(other, self.__class__):
			other = self.__class__(obj=other)
		return self.hexadecimal == other.hexadecimal

	def __hash__(self):
		return hash(self.hexadecimal)



from .BaseController import BaseController
from ._WorkerReport import WorkerReport


class SimpleController(BaseController):
	def do(self, echo=1):
		worker_id = self._generate_worker_id()
		report = WorkerReport(worker_id=worker_id)
		while self.to_do_count > 0:
			if echo:
				self.write(f'To-do: {self.to_do_count} - Doing: 1 - Done: {self.done_count} - Processed: 0       ')
			task = self.to_do_queue.popleft()
			self._tasks_status[task.id] = 'started'
			outcome = task.do(worker_id=worker_id)
			self.done_queue.append(outcome)
			self._tasks_status[task.id] = 'done'
			report.add_task_id(task_id=task.id)

		report.end()
		if echo:
			self.write(f'To-do: {self.to_do_count} - Doing: 0 - Done: {self.done_count} - Processed: 0       ')
		self.process_done_queue(echo=echo)


from .TimeMeasurement import TimeMeasurement
from .measure import measure
from .Timer import Timer
from .get_elapsed import get_elapsed, find_best_unit, convert

from pandas import DataFrame, concat
from numpy import median
from datetime import timedelta


class MeasurementSet:
	def __init__(self):
		self._measurements = {}
		self._start_times = {}
		self._end_times = {}

	def __str__(self):
		return '\n'.join([f'{name} - {measurement}' for name, measurement in self._measurements.items()])

	def __getstate__(self):
		return self._measurements, self._start_times, self._end_times

	def __setstate__(self, state):
		if len(state) == 3:
			self._measurements, self._start_times, self._end_times = state
		else:
			self._measurements = state
			self._start_times = self._end_times = {}

	def __hashkey__(self):
		return (self.__class__.__name__, self._measurements)

	@property
	def measurements(self):
		"""
		:rtype: dict[str, TimeMeasurement]
		"""
		return self._measurements

	@property
	def start_times(self):
		"""
		:rtype: dict[str, float or timedelta]
		"""
		return self._start_times

	@property
	def end_times(self):
		"""
		:rtype: dict[str, float or timedelta]
		"""
		return self._end_times

	def __add__(self, other):
		"""
		:type other: MeasurementSet
		:rtype: MeasurementSet
		"""
		result = self.__class__()
		for name in set(self.measurements.keys()).union(other.measurements.keys()):

			if name in self.measurements and name in other.measurements:
				result.measurements[name] = self.measurements[name] + other.measurements[name]

			elif name in self._measurements:
				result.measurements[name] = self.measurements[name].copy()

			else:
				result.measurements[name] = other.measurements[name].copy()

		return result

	def log(self, name, timer):
		"""
		:type timer: Timer
		"""
		if name in self._start_times:
			self._start_times[name].append(timer.last_start)
			self._end_times[name].append(timer.last_end)
		else:
			self._start_times[name] = [timer.last_start]
			self._end_times[name] = [timer.last_end]

	def add_measurement(self, name, timer):
		"""
		:param str name:
		:type timer: Timer
		"""
		self.log(name=name, timer=timer)

		if name in self.measurements:
			self.measurements[name] += TimeMeasurement(duration=timer.duration, unit=timer.unit)
		else:
			self.measurements[name] = TimeMeasurement(duration=timer.duration, unit=timer.unit)

	@property
	def performance_summary(self):
		"""
		:rtype: DataFrame
		"""
		names = []
		total_durations = []
		units = []
		counts = []
		mean_durations = []
		for name, measurement in self.measurements.items():
			names.append(name)
			total_durations.append(measurement.duration)
			units.append(measurement.unit)
			counts.append(measurement.count)
			mean_durations.append(measurement.mean_duration)

		data = DataFrame({
			'name': names, 'total_duration': total_durations, 'unit': units, 'count': counts
		})

		if all([unit == 'timedelta' for unit in units]):
			median_duration = median(mean_durations)
			best_unit = find_best_unit(delta=median_duration)
			data['unit'] = best_unit
			data['total_duration'] = data['total_duration'].apply(lambda duration: convert(delta=duration, to_unit=best_unit))

		data['mean_duration'] = data['total_duration'] / data['count']
		return data.sort_values('total_duration', ascending=False).reset_index(drop=True)

	@property
	def timestamps(self):
		"""
		:rtype: DataFrame
		"""
		dataframes = []
		for name in self._start_times.keys():
			dataframes.append(DataFrame({
				'name': name,
				'attempt': range(1, len(self._start_times[name]) + 1),
				'start': self._start_times[name],
				'end': self._end_times[name]
			}))
		result = concat(dataframes).reset_index(drop=True)
		all_times = list(result['start']) + list(result['end'])

		if len(all_times) == 0:
			return None
		else:
			min_time = min(all_times)
			differences = [t - min_time for t in all_times]
			best_unit = find_best_unit(delta=median(differences))
			first_time = all_times[0]

			result['standard_unit'] = best_unit
			result['standard_start'] = result['start'].apply(lambda t: get_elapsed(start=first_time, end=t, unit=best_unit))
			result['standard_end'] = result['end'].apply(lambda t: get_elapsed(start=first_time, end=t, unit=best_unit))
			return result

	def measure(self, function, name, unit='ms'):
		"""
		:param callable function: function to be timed
		:type name: str
		:type unit: str
		:rtype: callable
		"""
		return measure(measurement_set=self, function=function, name=name, unit=unit)



from pandas import DataFrame
from joblib import Parallel, delayed
from math import sqrt
import matplotlib.pyplot as plt

from ...time.progress import ProgressBar, iterate
from ...time import get_elapsed, get_now
from ...exceptions import ClusteringError

from .KMeans import KMeans


_CLUSTERING_OPTIMIZER_STATE = [
	'_min_k', '_max_k', '_n_jobs', '_models', '_kwargs', '_keep_models', '_distortions', '_inertias',
	'_silhouette_scores', '_timeout', '_external_parallelization', '_n_external_jobs', '_n_internal_jobs'
]


class ClusteringOptimizer:
	def __init__(
			self, min_k, max_k, n_jobs=-1, keep_models=False, timeout=None,
			external_parallelization=True, n_external_jobs=None, n_internal_jobs=None, **kwargs
	):
		self._min_k = min_k
		self._max_k = max_k
		self._n_jobs = n_jobs
		self._models = None
		self._kwargs = kwargs
		self._keep_models = keep_models
		self._distortions = None
		self._inertias = None
		self._silhouette_scores = None
		self._timeout = timeout
		self._external_parallelization = external_parallelization
		self._n_external_jobs = n_external_jobs
		self._n_internal_jobs = n_internal_jobs

	def __getstate__(self):
		return {
			name: getattr(self, name)
			for name in _CLUSTERING_OPTIMIZER_STATE
		}

	def __setstate__(self, state):
		for name, value in state.items():
			setattr(self, name, value)

	def fit(self, X, echo=1, raise_timeout=True):
		if self._keep_models:
			self._models = {}

		self._distortions = {}
		self._inertias = {}
		self._silhouette_scores = {}

		def get_scores(k, n_jobs):
			kmeans = KMeans(
				n_clusters=k, n_jobs=n_jobs, timeout=self._timeout, raise_timeout=raise_timeout, **self._kwargs
			)
			kmeans.fit(X=X)
			if kmeans.timedout:
				return k, None, None, None

			if self._keep_models:
				return k, kmeans, kmeans.inertia, kmeans.distortion, kmeans.silhouette_score

			else:
				return k, None, kmeans.inertia, kmeans.distortion, kmeans.silhouette_score

		if self._external_parallelization and self._n_jobs > 1:

			n_ks = self._max_k - self._min_k + 1
			n_external_jobs = self._n_external_jobs or min(n_ks, self._n_jobs)
			n_remaining_jobs = self._n_jobs - n_external_jobs
			n_internal_jobs = self._n_internal_jobs or max(1, int(round(n_remaining_jobs / n_ks)))
			print(f'n_ks = {n_ks} n_external_jobs = {n_external_jobs} n_internal_jobs = {n_internal_jobs}')

			processor = Parallel(n_jobs=n_external_jobs, backend='threading', require='sharedmem')
			result = processor(
				delayed(get_scores)(k=k, n_jobs=n_internal_jobs)
				for k in iterate(iterable=range(self._max_k, self._min_k - 1, -1), echo_items=True, echo=echo)
			)

			for k, model, inertia, distortion, silhouette_score in result:
				if inertia is not None or distortion is not None or silhouette_score is not None:
					self._distortions[k] = distortion
					self._inertias[k] = inertia
					self._silhouette_scores[k] = silhouette_score
					if self._keep_models:
						self._models[k] = model
				else:
					raise ClusteringError(f'Both inertia and distortion are None for k: {k}')

		else:

			progress_bar = ProgressBar(total=self._max_k - self._min_k + 1, echo=echo)
			error = ''

			elapsed = 0
			for k in range(self._max_k, self._min_k - 1, -1):
				start_time = get_now()
				progress_bar.show(amount=self._max_k - k, text=f'j={self._n_jobs}, k={k}{error}, t={round(elapsed)}')
				k, model, inertia, distortion, silhouette_score = get_scores(k=k, n_jobs=self._n_jobs)
				if inertia is not None or distortion is not None or silhouette_score is not None:
					if self._keep_models:
						self._models[k] = model
					self._distortions[k] = distortion
					self._inertias[k] = inertia
					self._silhouette_scores[k] = silhouette_score
				else:
					raise ClusteringError(f'Both inertia and distortion are None for k: {k}')
				elapsed = max(elapsed, get_elapsed(start=start_time, unit='s'))

			progress_bar.show(amount=progress_bar.total)

	def get_scores(self, by='silhouette_score'):
		"""
		:rtype: dict
		"""
		if by.startswith('silhouette'):
			return self._silhouette_scores.copy()
		elif by.startswith('inertia'):
			return self._inertias.copy()
		elif by.startswith('distortion'):
			return self._distortions.copy()
		else:
			raise ValueError(f'Unsupported scoring method: "{by}"')

	def plot(self, by='silhouette_score', size=(16, 8), title=None):
		scores = self.get_scores(by=by)
		if title is None:
			if by.startswith('silhouette'):
				title = 'Finding the Optimal k with the Silhouette Method'
			else:
				title = 'Finding the Optimal k with the Elbow Method'

		plt.figure(figsize=size)
		plt.plot(list(scores.keys()), list(scores.values()), 'bx-')
		plt.xlabel('k')
		plt.ylabel(by.replace('_', ' ').capitalize())
		plt.title(title)
		plt.show()

	def get_optimal_number_of_clusters(self, by='silhouette'):
		"""
		:param str by: can be 'inertia' or 'distortion'
		:rtype: int
		"""
		x1, x2 = min(self._silhouette_scores.keys()), max(self._silhouette_scores.keys())

		if by.startswith('silhouette'):
			return max(self._silhouette_scores, key=self._silhouette_scores.get)

		elif by.startswith('inertia'):
			y_dictionary = self._inertias

		elif by.startswith('distort'):
			y_dictionary = self._distortions

		else:
			raise ValueError(f'Method "{by}" is not supported!')

		# Elbow method
		y1, y2 = y_dictionary[x1], y_dictionary[x2]
		distances = []
		for x in range(x1, x2 + 1):
			y = y_dictionary[x]

			numerator = abs((y2 - y1) * x - (x2 - x1) * y + x2 * y1 - y2 * x1)
			denominator = sqrt((y2 - y1) ** 2 + (x2 - x1) ** 2)
			distances.append(numerator / denominator)

		return distances.index(max(distances)) + 2

	@property
	def optimal_number_of_clusters(self):
		"""
		:rtype: int
		"""

		scores = []
		for by in ['silhouette', 'inertia', 'distortion']:
			try:
				score = self.get_optimal_number_of_clusters(by=by)
				scores.append(score)
			except TypeError:
				continue

		return max(scores)

	@property
	def records(self):
		"""
		:rtype: list[dict]
		"""
		return [
			{'k': k, 'distortion': self._distortions[k], 'inertia': self._inertias[k]}
			for k in range(self._min_k, self._max_k + 1)
		]

	@property
	def data(self):
		"""
		:rtype: DataFrame
		"""
		return DataFrame.from_records(self.records)



from pandas import DataFrame
from ...collections import cross_lists


class Polynomial:
	def __init__(self, degree=2, sep='.'):
		self._degree = degree
		self._columns = None
		self._sep = sep

	@property
	def degree(self):
		"""
		:rtype: int
		"""
		return self._degree

	def fit(self, data):
		"""
		:type data: DataFrame
		"""
		if self._columns is not None:
			raise RuntimeError('polynomial has already been fit!')

		self._columns = list(data.columns)

	def get_columns(self, degree):
		combinations = cross_lists([self._columns] * degree)
		ordered_combinations = [sorted(x) for x in combinations]
		unique_combinations = []
		for combination in ordered_combinations:
			if combination not in unique_combinations:
				unique_combinations.append(combination)
		return {self.get_column_name(combination): combination for combination in unique_combinations}

	def get_column_name(self, combination):
		# count repetition of elements
		counts = {x: combination.count(x) for x in set(combination)}
		return self._sep.join([
			str(column) if counts[column] == 1 else f'{column}{counts[column]}'
			for column in sorted(set(combination))
		])

	def transform(self, data):
		"""
		:type data: DataFrame
		:rtype: DataFrame
		"""
		data = data.copy()
		for degree_minus_one in range(1, self._degree):
			for column_name, combination in self.get_columns(degree=degree_minus_one + 1).items():
				data[column_name] = 1
				for column in combination:
					data[column_name] = data[column_name] * data[column]
		return data

	def fit_transform(self, data):
		"""
		:type data: DataFrame
		:rtype: DataFrame
		"""
		self.fit(data=data)
		return self.transform(data=data)



from pandas import DataFrame


def balance_sample(data, columns, n=None, fraction=None, random_state=None):
	"""
	this method takes a sample of the data by reducing the imbalance between the values of columns
	:type data: DataFrame
	:type columns: str or list[str]
	:type n: int or NoneType
	:type fraction: float or NoneType
	:type random_state: NoneType or float or int
	"""

	if n is None:
		if fraction > 1:
			raise ValueError('fraction cannot be larger than 1')
		n = round(fraction * data.shape[0])
	if n > data.shape[0]:
		raise ValueError('n cannot be larger than the number of rows')

	n = min(n, data.shape[0])

	def add_index(x):
		x = x.sample(frac=1, random_state=random_state)
		x['__balance_sample_index__'] = range(x.shape[0])
		return x

	new_data = data.groupby(columns).apply(add_index)
	new_data.index = new_data.index.droplevel(0)
	new_data = new_data.sort_values('__balance_sample_index__').head(n)

	return new_data.drop(columns='__balance_sample_index__')



from builtins import property
from math import log10


class NumberPart:
	def __init__(self, value= 0, base=10, start=0, digits=None):

		if type(base) is not int and base is not None: raise TypeError("base should be int or None")
		if type(value) is not int: raise TypeError("value should be int")
		if type(start) is not int: raise TypeError("start should be int")

		self._base = base
		self._value = value
		self._start = start

		if digits is None:
			if self.base is None:
				self._digits=1
			else:
				self._digits=int(log10(self.base + self.start - 1))+1
		else:
			if type(digits) is not int: raise TypeError("digits should be int")
			self._digits = digits

		self._carry=0
		self.adjust()

	def __hashkey__(self):
		return (self.__class__.__name__, self.__getstate__())

	def __getstate__(self):
		return (self.base, self.value, self.start, self.digits, self.carry)

	def __setstate__(self, state):
		self._base, self._value, self._start, self._digits, self._carry = state

	def adjust(self):
		if self.base is not None:
			self._carry += (self._value - self._start) // self._base
			self._value = (self._value - self._start) % self._base + self._start

	@property
	def base(self):
		return self._base

	@property
	def value(self):
		return self._value

	@property
	def start(self):
		return self._start

	@property
	def carry(self):
		return self._carry

	@property
	def digits(self):
		return self._digits

	@value.setter
	def value(self, value):
		if type(value) is not int: raise TypeError("value should be int")
		self._value=value
		if self.base is not None:
			if self._value < self.start or self._value > self.base - self.start:
				self.adjust()

	def reset_carry(self):
		self._carry = 0

	def __repr__(self):
		if self.start==0:
			result = f"{self.value} carry:{self.carry} base:{self.base}"
		else:
			result = f"{self.value}(start={self.start}) carry:{self.carry} base:{self.base}"
		return result

	def __str__(self):
		if self.base is not None:
			return str(self.value + self.carry * self.base).zfill(self.digits)
		else:
			return str(self.value).zfill(self.digits)

	def copy(self):
		the_copy=self.__class__(value=self.value, base=self.base, start=self.start, digits=self.digits)
		the_copy._carry=self.carry
		return the_copy

	def same_class(self, other):
		if type(self) != type(other): return False
		if self.base != other.base: return False
		if self.start != other.start: return False
		if self.digits != other.digits: return False
		return True

	def __add__(self, other):
		if not self.same_class(other): raise TypeError('self and other are two different classes')
		result=self.__class__(value=self.value+other.value, base=self.base, start=self.start, digits=self.digits)
		result._carry += self.carry + other.carry
		return result

	def __sub__(self, other):
		if not self.same_class(other): raise TypeError('self and other are two different classes')
		result=self.__class__(value=self.value-other.value, base=self.base, start=self.start, digits=self.digits)
		result._carry += self.carry - other.carry
		return result

	def __mul__(self, other):
		result = self.__class__(value=self.value*other, base=self.base, start=self.start, digits=self.digits)
		result._carry = self._carry*other
		return result


# source: https://stackoverflow.com/questions/1653970/does-python-have-an-ordered-set
import collections


class OrderedSet(collections.OrderedDict, collections.MutableSet):
	def __init__(self, *args):
		super().__init__()
		if len(args) == 1:
			try:
				for element in args[0]:
					self.add(element)
			except TypeError:
				if args[0] is not None:
					self.add(args[0])
		else:
			for element in args:
				self.add(element)

	def update(self, *args, **kwargs):
		if kwargs:
			raise TypeError("update() takes no keyword arguments")

		for s in args:
			for e in s:
				self.add(e)

	def add(self, elem):
		self[elem] = None

	def discard(self, elem):
		self.pop(elem, None)

	def __le__(self, other):
		return all(e in other for e in self)

	def __lt__(self, other):
		return self <= other and self != other

	def __ge__(self, other):
		return all(e in self for e in other)

	def __gt__(self, other):
		return self >= other and self != other

	def __repr__(self):
		return 'OrderedSet([%s])' % (', '.join(map(repr, self.keys())))

	def __str__(self):
		return '{%s}' % (', '.join(map(repr, self.keys())))

	def __sub__(self, other):
		return super().__sub__(other)

	def __and__(self, other):
		return super().__and__(other)

	def __isub__(self, other):
		return super().__isub__(other)

	def __iand__(self, other):
		return super().__iand__(other)

	def __xor__(self, other):
		return super().__xor__(other)

	def __ixor__(self, other):
		return super().__ixor__(other)

	def __or__(self, other):
		return super().__or__(other)

	difference = __sub__
	difference_update = __isub__
	intersection = __and__
	intersection_update = __iand__
	issubset = __le__
	issuperset = __ge__
	symmetric_difference = __xor__
	symmetric_difference_update = __ixor__
	union = __or__



import sys
import threading


class SlytherinTimeout(TimeoutError):
	pass


try:
	import thread
except ImportError:
	import _thread as thread


def quit_function(fn_name):
	# print to stderr, unbuffered in Python 2.
	print('{0} took too long'.format(fn_name), file=sys.stderr)
	sys.stderr.flush()  # Python 3 stderr is likely buffered.
	thread.interrupt_main()  # raises KeyboardInterrupt


def exit_after(s):
	"""
	use as decorator to exit process if
	function takes longer than s seconds
	"""

	def outer(fn):
		def inner(*args, **kwargs):
			timer = threading.Timer(s, quit_function, args=[fn.__name__])
			timer.start()
			try:
				result = fn(*args, **kwargs)
			finally:
				timer.cancel()
			return result

		return inner

	return outer



from pandas import concat, DataFrame
from numpy import random

from .._Project import Project
from ...evaluation import evaluate_regression, evaluate_classification
from .._DataSlice import TrainingTestSlice
from ...validation import Scoreboard, TrainingTestContainer
from ._LearningTask import LearningTask
from ....collections.OrderedSet import OrderedSet


class LearningProject(Project):
	def __init__(
			self, name, y_column, problem_type, x_columns=None,
			time_unit='ms', evaluation_function=None, main_metric=None, lowest_is_best=None, best_score=None,
			scoreboard=None, processor=None
	):
		"""

		:type 	name: str
		:type 	y_column: str

		:type 	problem_type: str
		:param 	problem_type: either regression or classification

		:type 	x_columns: str

		:type 	time_unit: str
		:param 	time_unit: s, ms, etc.


		:type 	evaluation_function: callable
		:param 	evaluation_function: 	a function that gets predicted and actual and
										produces a dictionary of values such as {'rmse': ...} for regression
										or {'f1_score': ...} for classification

		:param 	main_metric: 	the main metric used for comparison, it should exist as one of the keys
								in the result produced by evaluation_function

		:param 	lowest_is_best: usually True for regression (unless a weird metric is used) and False for classification
		:param 	best_score: usually 0 for regression and 1 for classification

		:type 	scoreboard: Scoreboard
		:param 	scoreboard: a Scoreboard object that keeps score of all estimators, can be added later too
		"""
		super().__init__(name=name, time_unit=time_unit, processor=processor)
		self._estimators = {}
		self._training_test_slice_ids = OrderedSet()

		if problem_type.lower().startswith('reg'):
			self._problem_type = 'regression'

		elif problem_type.lower().startswith('class'):
			self._problem_type = 'classification'

		else:
			raise ValueError(f'problem_type: {problem_type} is not defined!')

		if evaluation_function is None:
			if self.problem_type == 'regression':
				self._evaluation_function = evaluate_regression
				main_metric = 'rmse' if main_metric is None else main_metric
				lowest_is_best = True if lowest_is_best is None else lowest_is_best
				best_score = 0 if best_score is None else best_score

			elif self.problem_type == 'classification':
				self._evaluation_function = evaluate_classification
				main_metric = 'f1_score' if main_metric is None else main_metric
				lowest_is_best = False if lowest_is_best is None else lowest_is_best
				best_score = 1 if best_score is None else best_score
		else:
			self._evaluation_function = evaluation_function
			if main_metric is None or lowest_is_best is None or best_score is None:
				raise ValueError(
					'main_metric, lowest_is_best, and best_score should be provided for evaluation_function')

		if y_column is None:
			raise ValueError('y_column should be provided')
		self._y_column = y_column
		self._x_columns = x_columns

		if scoreboard is None:
			scoreboard = Scoreboard(main_metric=main_metric, lowest_is_best=lowest_is_best, best_score=best_score)
		self._scoreboard = scoreboard
		self._all_tasks_produced = False

	def __repr__(self):
		lines = [
			super().__repr__(),
			f'estimators: {len(self.estimators)}',
			f'training_test_slice_id: {self._training_test_slice_ids}',
		]
		return '\n'.join(lines)

	def _generate_training_test_slice_id(self):
		return f'{self.name}_{len(self._training_test_slice_ids) + 1}'

	@property
	def y_column(self):
		"""
		:rtype: str
		"""
		return self._y_column

	@property
	def x_columns(self):
		"""
		:rtype list[str]
		"""
		return self._x_columns

	@x_columns.setter
	def x_columns(self, x_columns):
		if x_columns is None:
			raise ValueError('x_columns should be a list of strings!')
		if self.y_column in x_columns:
			raise KeyError(f'y_column "{self.y_column}" is among x_columns')

		if self._x_columns is None:
			self._x_columns = x_columns
		else:
			if set(self._x_columns) != set(x_columns):
				raise ValueError(f'x_columns cannot change!')

	def add_x_columns_from_data(self, data):
		if self.x_columns is None:
			self.x_columns = [column for column in data.column if column != self.y_column]

	def add_training_test_slice(
			self, training_test_slice, data=None, training_test_slice_id=None, overwrite=False,
			add_to_training_test_ids=True
	):
		"""
		:type training_test_slice_id: str
		:type training_test_slice: TrainingTestSlice
		:return:
		"""
		if training_test_slice_id is None:
			training_test_slice_id = self._generate_training_test_slice_id()

		if data is not None:
			self.processor.add_data(data_id=training_test_slice.data_id, data=data, overwrite=overwrite)

		columns = self.processor.get_obj(obj_type='columns', obj_id=training_test_slice.data_id)
		x_columns = [column for column in columns if column != self.y_column]
		self.x_columns = x_columns

		if add_to_training_test_ids:
			self._training_test_slice_ids.add(training_test_slice_id)

		self.processor.add_obj(
			obj_type='tts', obj_id=training_test_slice_id, obj=training_test_slice,
			overwrite=overwrite
		)

		if add_to_training_test_ids:
			self.scoreboard.add_training_test_id(training_test_id=training_test_slice_id)
			self._all_tasks_produced = False

	def add_training_test_container(self, container, training_test_slice_id=None, overwrite=False):
		"""
		:type training_test_slice_id: str
		:type container: TrainingTestContainer
		:type overwrite: bool
		"""
		if training_test_slice_id is None:
			training_test_slice_id = self._generate_training_test_slice_id()

		self.add_training_test_slice(
			training_test_slice_id=training_test_slice_id,
			data=container.data,
			overwrite=overwrite,
			training_test_slice=TrainingTestSlice(
				data_id=training_test_slice_id,
				training_indices=container.training_indices,
				test_indices=container.test_indices
			)
		)

	def add_training_test_data(self, training_data, test_data, training_test_slice_id=None, overwrite=False):
		if training_test_slice_id is None:
			training_test_slice_id = self._generate_training_test_slice_id()

		data = concat([training_data, test_data]).reset_index(drop=True)
		training_indices = list(range(training_data.shape[0]))
		test_indices = [x + training_data.shape[0] for x in range(test_data.shape[0])]
		tts = TrainingTestSlice(
			data_id=training_test_slice_id,
			training_indices=training_indices,
			test_indices=test_indices
		)
		self.add_training_test_slice(
			training_test_slice=tts, data=data, training_test_slice_id=training_test_slice_id,
			overwrite=overwrite
		)

	@staticmethod
	def _get_estimator_name(estimator_class):
		return estimator_class.__name__

	def add_estimator(self, estimator_class, estimator_id, estimator_arguments):
		if not isinstance(estimator_class, type):
			raise TypeError(f'estimator_class is of type {type(estimator_class)}')

		estimator_name = self._get_estimator_name(estimator_class)
		key = estimator_name, estimator_id
		self._estimators[key] = {'class': estimator_class, 'arguments': estimator_arguments}
		self.scoreboard.add_estimator(estimator_name=estimator_name, estimator_id=estimator_id)
		self._all_tasks_produced = False
		return estimator_name, estimator_id

	def add_estimator_repository(self, repository):
		"""
		:type repository: EstimatorRepository
		"""
		for dictionary in repository.estimator_dictionaries:
			estimator = dictionary['estimator']
			estimator_arguments = dictionary['estimator_arguments']
			estimator_id = dictionary['id']
			self.add_estimator(
				estimator_class=estimator,
				estimator_id=estimator_id,
				estimator_arguments=estimator_arguments
			)

	@property
	def estimators(self):
		"""
		:rtype: dict[(str, int), dict]
		"""
		return self._estimators

	@property
	def problem_type(self):
		return self._problem_type

	@property
	def evaluation_function(self):
		"""
		:rtype: callable
		"""
		return self._evaluation_function

	@property
	def scoreboard(self):
		"""
		:rtype: Scoreboard
		"""
		return self._scoreboard

	def produce_task(
			self, estimator_name, estimator_id, estimator_class, estimator_arguments,
			training_test_slice_id,
			ignore_error=False
	):
		"""

		:param estimator_name:
		:param estimator_id:
		:param estimator_class:
		:param estimator_arguments:
		:param training_test_slice_id:
		:param ignore_error:
		:rtype: LearningTask
		"""
		if self.x_columns is None:
			raise RuntimeError(f'x_columns is None')

		task = LearningTask(
			project_name=self.name, estimator_class=estimator_class,
			estimator_name=estimator_name, estimator_id=estimator_id,
			estimator_arguments=estimator_arguments,
			training_test_slice_id=training_test_slice_id,
			y_column=self.y_column, x_columns=self.x_columns,
			evaluation_function=self.evaluation_function
		)
		if self.contains_task(task_id=task.id):
			if ignore_error:
				return None
			else:
				raise RuntimeError(f'task {task} already exists in project {self.name}')
		return task

	def produce_tasks(self, ignore_error=False, echo=True):
		task_count = 0
		for training_test_slice_id in self._training_test_slice_ids:
			for estimator_name_and_id, estimator_class_and_arguments in self._estimators.items():
				estimator_name, estimator_id = estimator_name_and_id
				task = self.produce_task(
					estimator_name=estimator_name, estimator_id=estimator_id,
					estimator_class=estimator_class_and_arguments['class'],
					estimator_arguments=estimator_class_and_arguments['arguments'],
					training_test_slice_id=training_test_slice_id,
					ignore_error=ignore_error
				)
				if task is not None:
					task_count += 1

					self._pre_to_do[task.id] = task

		if echo:
			print(f'{task_count} tasks produced for project {self.name}')
		self._all_tasks_produced = True

	def _get_new_to_do_list(self, num_tasks=1, method='upper_bound', random_state=None, echo=True):
		if num_tasks > self.new_count:
			raise ValueError(f'num_tasks {num_tasks} is too large! There are only {self.new_count} available')
		if len(self._pre_to_do) == 0:
			if echo:
				print('no new tasks to fill the to-do list')
			return None

		data = DataFrame.from_records([
			{
				'estimator_name': task.estimator_name, 'estimator_id': task.estimator_id,
				'training_test_id': task.training_test_id,
				'task_id': task.id
			}
			for task in self._pre_to_do.values()
		])
		if method == 'upper_bound':
			mean_per_data = self.scoreboard.mean_score_per_data.rename(columns={'score': 'data_score'})
			data = data.merge(mean_per_data, on='training_test_id', how='left')

			best_per_estimator = self.scoreboard.best_possible_score_per_estimator
			best_per_estimator = best_per_estimator.rename(columns={'score': 'estimator_score'})
			data = data.merge(best_per_estimator, on=['estimator_name', 'estimator_id'], how='left')
			if random_state is not None:
				random.seed(random_state)

			# we want to try the estimators that have the best shot
			# also we want to avoid repeating the same estimator
			# (hence estimator_repetition counts the number of
			# times an estimator appears, we put estimator_repetition first
			# because it is useless if it comes after estimator_score
			# and we want to prioritize data sets that have the worst outcome, sooner rather than later
			# on top of all that, when everything is equal, randomize

			data['random'] = random.uniform(size=data.shape[0])
			if data['estimator_score'].isnull().values.any():
				raise RuntimeError('there are nulls among estimator_scores')

			ascending = [self.scoreboard.lowest_is_best, not self.scoreboard.lowest_is_best, True]

			data.sort_values(
				by=['estimator_score', 'data_score', 'random'],
				ascending=ascending,
				na_position='first',
				inplace=True
			)

			data['estimator_repetition'] = data.groupby(['estimator_name', 'estimator_id']).cumcount()
			data.sort_values(
				by=['estimator_repetition', 'estimator_score', 'data_score', 'random'],
				ascending=[True] + ascending,
				na_position='first',
				inplace=True
			)

		elif method == 'random':
			data = data.sample(frac=1, random_state=random_state)

		else:
			raise ValueError(f'method {method} is unknown!')

		return [row['task_id'] for index, row in data.head(num_tasks).iterrows()]

	def fill_to_do_list(self, num_tasks=1, method='upper_bound', random_state=None, echo=True):
		task_ids = self._get_new_to_do_list(num_tasks=num_tasks, method=method, random_state=random_state, echo=echo)

		filled_count = 0
		for task_id in task_ids:
			self._take_from_pre_and_add_to_to_do(task_id=task_id)
			filled_count += 1

		if echo:
			print(f'{filled_count} to-do tasks added to project {self.name}')

	def process(self, task):
		"""
		:type task: LearningTask
		"""
		if task.has_error():
			for error, trace in task.errors:
				print(f'trace: {trace}')
			raise task.errors[0][0]
		if task.status != 'done':
			raise RuntimeError(f'task is not done. Task status is {task.status}')

		if not isinstance(task.evaluation, dict):
			raise TypeError(f'evaluation is of type {type(task.evaluation)}')
		self._scoreboard.add_task_score(task=task)

	def get_best_estimators(self, num_estimators=1):
		scores = self.scoreboard.mean_score_per_estimator.sort_values(
			'score', ascending=self.scoreboard.lowest_is_best
		)
		if scores.shape[0] < num_estimators:
			raise RuntimeError(f'there are only {scores.shape[0]} estimators to choose from!')
		result = []
		for row_num, row in scores.head(num_estimators).iterrows():
			estimator_name = row['estimator_name']
			estimator_id = row['estimator_id']
			score = row['score']
			dictionary = self.estimators[(estimator_name, estimator_id)]
			parameters = {
				'name': estimator_name, 'id': estimator_id, 'class': dictionary['class'], 
				'arguments': dictionary['arguments'], 'score': score
			}
			result.append(parameters)
		return result

	def get_best_estimator(self):
		return self.get_best_estimators(num_estimators=1)[0]



from sklearn.metrics import mean_squared_error, mean_absolute_error
from math import sqrt
import numpy as np
from warnings import warn, catch_warnings, simplefilter


def rmse(actual, predicted):
	actual = np.array(actual)
	predicted = np.array(predicted)
	return sqrt(mean_squared_error(actual, predicted))


root_mean_squared_error = rmse


def mape(actual, predicted, ignore_warnings=True):
	actual = np.array(actual)
	predicted = np.array(predicted)
	if ignore_warnings:
		with catch_warnings():
			simplefilter('ignore')
			result = np.mean(np.abs((actual - predicted) / ((actual + predicted)/2))) * 100
	else:
		result = np.mean(np.abs((actual - predicted) / ((actual + predicted) / 2))) * 100
	return result


# symmetric mean absolute percentage error
def smape(actual, predicted, ignore_warnings=True):
	actual = np.array(actual)
	predicted = np.array(predicted)
	if ignore_warnings:
		with catch_warnings():
			simplefilter('ignore')

			result = np.mean(np.abs(actual - predicted) / ((np.abs(actual) + np.abs(predicted)) / 2)) * 100
	else:
		result = np.mean(np.abs(actual - predicted) / ((np.abs(actual) + np.abs(predicted)) / 2)) * 100
	return result


mean_absolute_percentage_error = mape
symmetric_mean_absolute_percentage_error = smape
mae = mean_absolute_error


def nmae(actual, predicted, ignore_warnings=True):
	actual = np.array(actual)
	predicted = np.array(predicted)
	if ignore_warnings:
		with catch_warnings():
			simplefilter('ignore')
			result = mae(actual, predicted)/(actual.max() - actual.min())
	else:
		result = mae(actual, predicted) / (actual.max() - actual.min())
	return result


normalized_mean_absolute_error = nmae


def evaluate_regression(actual, predicted, raise_error=True, ignore_warnings=True):
	actual = np.array(actual)
	predicted = np.array(predicted)

	# remove nas
	predicted = predicted[np.logical_not(np.isnan(actual))]
	actual = actual[np.logical_not(np.isnan(actual))]

	_mae = _rmse = _mape = _nmae = _smape = None
	try:
		_mae = mae(actual, predicted)
		if ignore_warnings:
			with catch_warnings():
				simplefilter('ignore')
				_nmae = _mae / (actual.max() - actual.min())
		else:
			_nmae = _mae / (actual.max() - actual.min())
		_rmse = rmse(actual=actual, predicted=predicted)
		_mape = mape(actual=actual, predicted=predicted, ignore_warnings=ignore_warnings)
		_smape = smape(actual=actual, predicted=predicted, ignore_warnings=ignore_warnings)
	except Exception as e:
		if raise_error:
			raise e
		else:
			message = f'ignoring error in evaluate_regression: {e}'
			print(message)
			warn(message)
	return {'nmae': _nmae, 'rmse': _rmse, 'mae': _mae, 'mape': _mape, 'smape': _smape}



from time import sleep
from ._WorkerReport import WorkerReport


def do_task(
		to_do_queue, done_queue, workers_doing, tasks_status, keep_workers_alive,
		worker_reports, worker_id,
		cpu_usage, max_cpu_count,
		empty_count_limit, max_sleep_time,
		sleep_time=0.1, echo=0
):
	"""
	:type to_do_queue: multiprocessing.Queue[Task]
	:type done_queue: multiprocessing.Queue[Outcome]
	:type workers_doing: dict
	:type tasks_status: dict
	:type keep_workers_alive: multiprocessing.Value
	:type worker_reports: dict
	:type worker_id: str or int
	:type cpu_usage: multiprocessing.Value
	:type max_cpu_count: int
	:type empty_count_limit: int
	:type max_sleep_time: float
	:type sleep_time: float
	:type echo: bool or int
	:rtype:
	"""

	report = WorkerReport(worker_id=worker_id)

	def print_if_echo(x):
		if echo:
			print(x)

	empty_count = 0
	# sleep(sleep_time)
	while max_cpu_count > cpu_usage.value:
		try:
			print_if_echo(f'{worker_id} getting task')
			task = to_do_queue.get_nowait()
			empty_count = 0
		except Exception:
			print_if_echo(f'{worker_id} queue empty')
			empty_count += 1
			if keep_workers_alive.value == 0:
				break
			if empty_count_limit is not None and empty_count >= empty_count_limit:
				break
			else:
				sleep(min(sleep_time * 2 ** empty_count, max_sleep_time))
		else:
			print_if_echo(f'{worker_id} got task {task.id}')

			tasks_status[task.id] = 'started'
			workers_doing[worker_id] = task.id

			print_if_echo(f'{worker_id} doing task {task.id}')
			cpu_usage.value += task.cpu_count
			outcome = task.do(worker_id=worker_id)
			cpu_usage.value -= task.cpu_count
			print_if_echo(f'{worker_id} did task {task.id}')

			report.add_task_id(task_id=task.id)

			print_if_echo(f'{worker_id} putting result of {task.id}')
			done_queue.put(outcome)

			workers_doing[worker_id] = None
			tasks_status[task.id] = 'done'

			print_if_echo(f'{worker_id} waiting')
			sleep(sleep_time)

	print_if_echo(f'{worker_id} out of the loop')
	report.end()
	print_if_echo(f'{worker_id} putting report')
	worker_reports[worker_id] = report
	return True



from pandas import DataFrame


def create_data_from_combinations(**kwargs):
	"""
	:rtype: DataFrame
	"""
	data_frames = [
		DataFrame({name: list(set(values))})
		for name, values in kwargs.items()
	]
	result = None
	for data in data_frames:
		data['_id_'] = 1
		if result is None:
			result = data
		else:
			result = result.merge(right=data, on='_id_', how='left')
	return result.drop(columns='_id_')



class Measurement:
	def __init__(self, x, result, elapsed, timeout_error, other_error):
		"""
		:type x: dict
		:type result: object
		:type elapsed: float
		:type timeout_error: bool
		"""
		if not isinstance(x, dict):
			raise TypeError('x should be a dictionary!')
		self._x = x
		self._result = result
		self._elapsed_time = elapsed
		self._timeout_error = timeout_error
		self._other_error = other_error
		self._weight = None

	@property
	def weight(self):
		return self._weight

	@property
	def timeout_error(self):
		return self._timeout_error

	@property
	def other_error(self):
		return self._other_error

	def __lt__(self, other):
		return self.elapsed_time < other.elapsed_time

	def __le__(self, other):
		return self.elapsed_time <= other.elapsed_time

	def __eq__(self, other):
		return self.elapsed_time == other.elapsed_time

	def __gt__(self, other):
		return self.elapsed_time > other.elapsed_time

	def __ge__(self, other):
		return self.elapsed_time >= other.elapsed_time

	def __ne__(self, other):
		return self.elapsed_time != other.elapsed_time

	@property
	def x(self):
		"""
		:rtype: dict
		"""
		return self._x

	@property
	def elapsed_time(self):
		return self._elapsed_time

	@property
	def dictionary(self):
		if isinstance(self._result, dict) and 'elapsed' not in self._result and 'x' not in self._result:
			return dict(
				x=self._x, time=self.elapsed_time, timeout_error=self._timeout_error, other_error=self._other_error,
				**self._result
			)
		else:
			return {
				**self._x, 'time': self.elapsed_time, 'result': self._result,
				'timeout_error': self._timeout_error, 'other_error': self._other_error
			}



from sklearn.linear_model import LinearRegression
from sklearn.tree import DecisionTreeClassifier
from pandas import DataFrame, concat
from pandas.api.types import is_numeric_dtype
from .exceptions import NoColumnsError, ImputerNotFittedError
from ._find_good_columns import find_best_coverage_combination
ROW_NUM_COL = '_imputer_row_number_'


class SingleColumnImputer:
	def __init__(self, estimator, column, use_columns, column_type=None):
		"""
		:type estimator: LinearRegression or DecisionTreeClassifier or HyperModel
		:type column: str
		"""

		self._estimator = estimator
		self._imputed_column = column
		self._helper_columns = None
		self._coverage = None
		self._use_columns = use_columns
		self._fitted = None
		self._column_type = column_type
		if self._estimator is None:
			raise RuntimeError(f'no estimator provided!')

	def fit(self, X):
		"""
		:type X: DataFrame
		"""
		d = find_best_coverage_combination(data=X, column=self._imputed_column)
		self._helper_columns = [column for column in d['combination'] if column in self._use_columns]
		self._coverage = d

		data = X[self._helper_columns + [self._imputed_column]]

		training_data = data.dropna(axis=0)
		training_X = training_data[self._helper_columns]
		training_y = training_data[self._imputed_column]

		if training_X.shape[1] == 0:
			raise NoColumnsError('X has no columns!')
		elif training_X.shape[0] == 0:
			raise NoColumnsError('X has no rows!')

		if self._estimator is None:
			raise RuntimeError('no estimator')

		try:
			self._estimator.fit(training_X, training_y)
		except:
			display(training_X.head(), training_y.head())
			raise

		self._fitted = True
		if self._column_type is None:
			if is_numeric_dtype(training_y):
				self._column_type = 'numerical'
			else:
				self._column_type = 'nonnumerical'

	@property
	def type(self):
		if self._column_type.startswith('numeric'):
			return 'regressor'
		else:
			return 'classifier'

	def impute_column(self, data):
		missing = data[data[self._imputed_column].isna()].copy()
		not_missing = data[data[self._imputed_column].notna()]
		missing[self._imputed_column] = self._estimator.predict(missing[self._helper_columns])

		all_data = concat([missing, not_missing]).sort_values(ROW_NUM_COL)
		return all_data[self._imputed_column]

	def transform(self, X):
		"""
		:type X: DataFrame
		:rtype: DataFrame
		"""
		if not self._fitted:
			raise ImputerNotFittedError('imputer is not fitted yet!')

		data = X.copy()

		data[ROW_NUM_COL] = range(data.shape[0])
		data[self._imputed_column] = self.impute_column(data=data)
		return data.drop(columns=ROW_NUM_COL)

	def fit_transform(self, X):
		self.fit(X)
		return self.transform(X)



from builtins import property
from .NumberPart import NumberPart


class FlexibleNumber:
	def __init__(self, parts=None, labels=None, sep='-'):
		self._parts = parts or []
		self._sep = sep

		if labels is None:
			labels = list(reversed(range(len(parts))))
		self._labels = labels

		if len(parts) != len(labels):
			raise ValueError(f"parts and labels should be the same length but they are {len(parts)} and {len(labels)} respectively.")

		if len(labels) != len(set(labels)):
			raise ValueError("labels are not unique")

	def __hashkey__(self):
		return (self.__class__.__name__, self.__getstate__())

	def __getstate__(self):
		return (self._parts, self._sep, self._labels)

	def __setstate__(self, state):
		self._parts, self._sep, self._labels = state

	@property
	def parts(self):
		return self._parts

	@property
	def labels(self):
		return self._labels

	def len(self):
		return len(self._parts)

	def adjust(self):
		for index in reversed(range(self.len())):
			this_part = self.parts[index]
			if index > 0: # there is another element to the left
				left_part = self.parts[index - 1]
				left_part.value += this_part.carry
				this_part.reset_carry()

	def __str__(self):
		return self._sep.join([str(part) for part in self.parts])

	def __repr__(self):
		return ' | '.join([repr(part) for part in self.parts])

	def get(self, label):
		"""
		:type label: str
		:rtype: NumberPart
		"""
		return self.parts[self.labels.index(label)]

	def get_value(self, label):
		"""
		:type label: str
		:rtype: int
		"""
		return self.get(label=label).value

	def set(self, value, label):
		self.parts[self.labels.index(label)].value = value
		self.adjust()

	def add(self, value=0, label=None):
		if label is None:
			label = self.labels[-1]
		self.parts[self.labels.index(label)].value += value
		self.adjust()

	def get_total(self, label = None):
		if label is None:
			last_index = self.len() - 1
		else:
			last_index = self.labels.index(label)

		total = 0
		for i in range(last_index):
			total += self.parts[i].value + self.parts[i].carry
			total *= self.parts[i+1].base
		total += self.parts[last_index].value + self.parts[last_index].carry

		return total

	def same_class(self, other):
		if type(self) != type(other): return False
		if self.labels != other.labels: return False
		for index in range(self.len()):
			if not self.parts[index].same_class(other.parts[index]): return False
		return True

	def __hash__(self):
		return self.get_total()


	def __eq__(self, other):
		if not self.same_class(other):
			return False
		else:
			return self.get_total() == other.get_total()

	def __ne__(self, other):
		return not self.__eq__(other)

	def __gt__(self, other):
		if not self.same_class(other):
			raise TypeError("the other object is a different class or different FlexibleNumber")
		return self.get_total() > other.get_total()

	def __le__(self, other):
		return not self.__gt__(other)

	def __lt__(self, other):
		if not self.same_class(other):
			raise TypeError("the other object is a different class or different FlexibleNumber")
		return self.get_total() < other.get_total()

	def __ge__(self, other):
		return not self.__lt__(other)




	def copy(self):
		parts = [part.copy() for part in self.parts]
		return self.__class__(parts=parts, labels=self.labels, sep=self._sep)

	def __add__(self, other):
		if not self.same_class(other):
			raise TypeError("the other object is a different class or different FlexibleNumber")

		parts = []
		for i in range(self.len()):
			part = self.parts[i] + other.parts[i]
			part.explore()
			parts.append(part)

		return  self.__class__(parts=parts, labels=self.labels, sep=self._sep)

	def __sub__(self, other):
		if not self.same_class(other):
			raise TypeError("the other object is a different class or different FlexibleNumber")

		parts = []
		for i in range(self.len()):
			part = self.parts[i] - other.parts[i]
			part.explore()
			parts.append(part)

		return  self.__class__(parts=parts, labels=self.labels, sep=self._sep)





from .add_time import add_time, add_days, add_weeks, add_months, add_quarters, add_years
from .convert_string_to_date import convert_string_to_date
from .get_date_part import get_date_part, get_monthname, get_weekday, get_weekdayname, yearmonth_to_date
from .get_month import get_month_name, get_month_num, get_months_list
from .get_number_of_days import get_number_of_days
from .get_quarter import get_quarter
from .get_today import get_today, get_today_str, get_date
from .parse_date import parse_date, find_date_format, find_date_and_format
from .YearMonth import YearMonth
from .YearQuarter import YearQuarter
from .get_first_day_of_month import get_first_day_of_month, get_last_day_of_month
from .get_date_mean import get_date_mean
from .get_week_number import get_week_number



from re import sub


def split_and_strip(string, sep=','):
	if type(string) is str:
		return sub(pattern=r'\s*{0}\s*'.format(sep), repl=sep, string=string.strip()).split(sep=sep)
	else:
		return string








from pandas.api.types import is_numeric_dtype
from ...collections import get_intersection
from ...multiprocessing import JobController


class Imputer:
	def __init__(
			self, estimators, cross_validation, ignore_columns=None, helper_columns=None, impute_columns=None,
			main_metric='rmse', impute_categorical=False,
			controller=None
	):
		"""
		:type estimators: LinearRegression or DecisionTreeClassifier or list
		:type ignore_columns: NoneType or str or list[str]
		:type helper_columns: NoneType or str or list[str]
		:type impute_columns: NoneType or str or list[str]
		:type cross_validation: CrossValidation or NoneType
		:type main_metric: str
		:type impute_categorical: bool
		"""
		if ignore_columns is None:
			ignore_columns = []
		elif isinstance(ignore_columns, str):
			ignore_columns = [ignore_columns]
		elif not isinstance(ignore_columns, list):
			raise TypeError(f'ignore_columns cannot be of type {type(ignore_columns)}')

		self._ignore_columns = ignore_columns

		if helper_columns is not None:
			if isinstance(helper_columns, str):
				helper_columns = [helper_columns]

			if len(get_intersection(helper_columns, ignore_columns)) > 0:
				raise ValueError(f'helper_columns and ignore_columns intersect: {get_intersection(helper_columns, ignore_columns)}')

		self._helper_columns = helper_columns

		if impute_columns is not None:
			if isinstance(impute_columns, str):
				impute_columns = [impute_columns]

			if len(get_intersection(impute_columns, ignore_columns)) > 0:
				raise ValueError(f'impute_columns and ignorre_columns intersect: {get_intersection(impute_columns, ignore_columns)}')

		self._impute_columns = impute_columns

		if not isinstance(estimators, (list, tuple)):
			estimators = [estimators]

		self._estimators = estimators
		self._cross_validation = cross_validation
		self._main_metric = main_metric.lower()
		self._impute_categorical = impute_categorical
		self._column_imputers = {}

		if controller is None:
			controller = JobController()
		self._controller = controller

	def _get_ignore_columns(self, data):
		return [col for col in data.columns if col in self._ignore_columns]

	def _get_helper_columns(self, data, ignore_columns):
		if self._helper_columns is not None:
			return self._helper_columns
		else:
			return [col for col in data.columns if col not in ignore_columns]

	def _get_impute_columns(self, data, ignore_columns):
		if self._impute_columns is not None:
			return self._impute_columns
		else:
			impute_columns = [
				col for col in data.columns
				if col not in ignore_columns and data[col].isnull().values.any()
			]
			if self._impute_categorical:
				return impute_columns
			else:
				return [col for col in impute_columns if is_numeric_dtype(data[col])]

	def fit(self, X, echo=1):
		ignore_columns = self._get_ignore_columns(data=X)
		helper_columns = self._get_helper_columns(data=X, ignore_columns=ignore_columns)
		impute_columns = self._get_impute_columns(data=X, ignore_columns=ignore_columns)

		def _evaluate_estimator(impute_column, fold_num, )


		for


from sklearn.linear_model import LinearRegression, LogisticRegression
from copy import deepcopy
from sklearn.base import is_classifier, is_regressor
from ._Result import TrainingResult, TestResult
from ...exceptions import EstimatorNotEvaluatedError, ComparisonError
from ...time.time import Job


class Evaluation(Job):
	def __init__(self, estimator, training_x, training_y, test_x, test_y, estimator_id, problem_type, main_metric):
		super().__init__()
		self._training_result = None
		self._test_result = None
		self._estimator = estimator
		self._estimator_id = estimator_id
		self._training_x = training_x
		self._training_y = training_y
		self._test_x = test_x
		self._test_y = test_y
		self._evaluated = False

		if problem_type is None:
			if is_regressor(estimator):
				problem_type = 'regression'
			elif is_classifier(estimator):
				problem_type = 'classification'
			else:
				raise ValueError('problem type is not provided!')

		if problem_type.lower().startswith('regres'):
			problem_type = 'regression'
		elif problem_type.lower().startswith('class'):
			problem_type = 'classification'

		if main_metric is None:
			if problem_type == 'regression':
				main_metric = 'smape'
			elif problem_type == 'classification':
				main_metric = 'f1_score'
			else:
				raise ValueError('main metric should be provided!')

		self._problem_type = problem_type
		self._main_metric = main_metric

	@property
	def estimator_name(self):
		return str(self._estimator.__class__.__name__)

	@property
	def estimator_id(self):
		return self._estimator_id

	def fit(self):
		"""
		:rtype: LinearRegression or LogisticRegression
		"""
		estimator = deepcopy(self._estimator)
		estimator.fit(X=self._training_x, y=self._training_y)
		return estimator

	def evaluate(self, estimator):
		"""
		:type estimator: LinearRegression or LogisticRegression
		"""
		test_predicted = estimator.predict(self._test_x)
		test_actual = self._test_y

		training_predicted = estimator.predict(self._training_x)
		training_actual = self._training_y

		try:
			test_probabilities = estimator.predict_proba(self._test_x)
			training_probabilities = estimator.predict_proba(self._training_x)
		except AttributeError:
			test_probabilities = None
			training_probabilities = None

		self._test_result = TestResult(
			actual=test_actual, predicted=test_predicted, probabilities=test_probabilities,
			problem_type=self._problem_type
		)
		self._training_result = TrainingResult(
			actual=training_actual, predicted=training_predicted, probabilities=training_probabilities,
			problem_type=self._problem_type
		)
		self._evaluated = True

	def do(self):
		"""
		:rtype: LinearRegression or LogisticRegression
		"""
		self.start()
		estimator = self.fit()
		self.evaluate(estimator=estimator)
		self.end()
		return estimator

	def is_evaluated(self):
		return self._evaluated

	@property
	def training(self):
		"""
		:rtype: TrainingResult
		"""
		if self.is_evaluated():
			return self._training_result
		else:
			raise EstimatorNotEvaluatedError(f'estimator "{self.estimator_name}" is not evaluated yet!')

	@property
	def test(self):
		"""
		:rtype: TestResult
		"""
		if self.is_evaluated():
			return self._test_result
		else:
			raise EstimatorNotEvaluatedError(f'estimator "{self.estimator_name}" is not evaluated yet!')

	@property
	def performance(self):
		return self.test.performance[self._main_metric]

	def _check_comparability(self, other):
		"""
		:type other: Evaluation
		"""
		if not isinstance(other, self.__class__):
			raise TypeError(f'{other} is of type {type(other)}!')
		if self._problem_type != other._problem_type:
			raise ComparisonError(
				f'Cannot compare a problem type of "{self._problem_type}" with "{other._problem_type}"!'
			)
		if self._main_metric != other._main_metric:
			raise ComparisonError(
				f'Main metric is ambiguous!'
			)

	def __gt__(self, other):
		"""
		:type other: Evaluation
		"""
		self._check_comparability(other)
		if self._problem_type.lower()[0] == 'r':  #  regression:
			return self.performance < other.performance
		else:
			return self.performance > other.performance

	def __eq__(self, other):
		"""
		:type other: Evaluation
		"""
		self._check_comparability(other)
		return self.performance == other.performance

	def __ge__(self, other):
		return self == other or self > other

	def __lt__(self, other):
		return other > self

	def __le__(self, other):
		return other >= self

	def __ne__(self, other):
		return not (self == other)



from sklearn.linear_model import LinearRegression
from pandas import DataFrame

from ...time.progress import ProgressBar


class Ensemble:
	def __init__(self, models, weights=None, problem_type='regression', echo=1):
		"""
		:type models: list[LinearRegression]
		:type weights: NoneType of list[float]
		:type problem_type: either 'regression' or 'classification'
		"""
		self._problem_type = problem_type
		self._models = models
		if weights is None:
			self._weights = [1 for model in models]

		else:
			if len(weights) != len(models):
				raise ValueError(f'{len(models)} models need {len(models)} weights!')
			self._weights = weights
		self._fitted = False
		self._echo = echo

	def __repr__(self):
		result = ''
		count = 0
		for weight, model in zip(self._weights, self._models):
			if count >= 3:
				break
			if result != '':
				result += '\n\n'
			result += f'weight: {weight}\n'
			result += repr(model)
			count += 1
		return result

	def fit(self, X, y):
		num_models = len(self._models)
		progress_bar = ProgressBar(total=num_models, echo=self._echo)
		progress = 0
		for model in self._models:
			progress_bar.show(amount=progress, text=f'models fitted: {progress}/{num_models}')
			model.fit(X, y)
			progress += 1
		self._fitted = True
		progress_bar.show(amount=progress, text=f'models fitted: {progress}/{num_models}')

	def predict(self, X):
		if self._fitted is False:
			raise RuntimeError('models are not fitted yet!')

		if self._problem_type.lower()[0] == 'r':
			predictions = [model.predict(X) for model in self._models]

			weighted_sum = None
			for result, weight in zip(predictions, self._weights):
				if weighted_sum is None:
					weighted_sum = result * weight

				else:
					weighted_sum = weighted_sum + (result * weight)

			weighted_average = weighted_sum / sum(self._weights)
			return weighted_average

		else:

			predictions = {f'model_{i}': model.predict(X) for i, model in enumerate(self._models)}
			weights = {f'model_{i}': weight for i, weight in enumerate(self._weights)}
			votes = None

			for model_name, prediction in predictions.items():
				if votes is None:
					votes = DataFrame({model_name: list(prediction)})

				else:
					votes[model_name] = prediction

			def _get_weighted_vote(row):
				result = {}
				for model_name, w in weights.items():
					vote = row[model_name]
					if vote in result:
						result[vote] += w
					else:
						result[vote] = w

				max_score = 0
				winner = None
				for model_name, score in result.items():
					if score > max_score:
						winner = model_name
						max_score = score
				return winner
			return votes.apply(_get_weighted_vote, axis=1)



from collections import OrderedDict
from ...time.progress import ProgressBar
from ._TimeEstimate import TimeEstimate, MissingTimeEstimate
from ._Task import Task
import warnings


class Project:
	def __init__(self, name, processor, time_unit='ms'):
		"""
		:type name: str
		:type time_unit: str
		"""
		self._name = name
		self._time_unit = time_unit
		self._time_estimates = {}
		self._pre_to_do = OrderedDict()
		self._to_do = OrderedDict()

		self._being_done_ids = set()

		self._done = OrderedDict()
		self._processor = processor
		processor.add_project(project=self)

	def __repr__(self):
		lines = [
			f'Name: {self.name}',
			'',
			f'tasks: {self.task_count} (to-do: {self.to_do_count}, being done: {self.being_done_count}, done: {self.done_count})'
		]
		return '\n'.join(lines)

	def __str__(self):
		return str(self.name)

	@property
	def name(self):
		return self._name

	@property
	def time_estimates(self):
		"""
		:rtype: dict[str, TimeEstimate]
		"""
		return self._time_estimates

	def add_time_estimate(self, task):
		"""
		:type task: TrainingTestTask
		"""
		if task.estimator_name not in self.time_estimates:
			self.time_estimates[task.time_estimate_id] = TimeEstimate()
		self.time_estimates[task.time_estimate_id].append(task.get_elapsed(unit=self._time_unit))

	def get_time_estimate(self, task):
		if task.project_name != self.name:
			raise RuntimeError(f'task.problem_id = {task.project_name} does not match problem_id = {self.name}')

		if task.is_done():
			return task.get_elapsed(unit=self._time_unit)

		if task.estimator_name in self.time_estimates:
			return self.time_estimates[task.time_estimate_id].get_mean()

		elif len(self.time_estimates) > 0:
			total = 0
			count = 0
			for estimate in self.time_estimates.values():
				total += estimate.mean
				count += 1
			return total / count

		else:
			return MissingTimeEstimate()

	@property
	def processor(self):
		"""
		:rtype: atlantis.ds.parallel_computing.Processor
		"""
		return self._processor

	@property
	def new_count(self):
		return len(self._pre_to_do)

	@property
	def to_do_count(self):
		return len(self._to_do)

	@property
	def being_done_count(self):
		return len(self._being_done_ids)

	@property
	def done_count(self):
		return len(self._done)

	@property
	def task_count(self):
		return self.new_count + self.to_do_count + self.being_done_count + self.done_count

	def contains_task(self, task_id):
		if isinstance(task_id, Task):
			raise TypeError('task_id cannot be of type Task!')
		return task_id in self._pre_to_do or task_id in self._to_do or task_id in self._being_done_ids or task_id in self._done

	def _take_from_pre_and_add_to_to_do(self, task_id):
		if task_id in self._pre_to_do:
			task = self._pre_to_do.pop(key=task_id)
			self._to_do[task_id] = task
		else:
			raise KeyError(f'task_id {task_id} does not exist in pre-to-do')

	def pop_to_do(self):
		"""
		:rtype: Task
		"""
		task_id, task = self._to_do.popitem(0)
		if task.id in self._being_done_ids:
			self._to_do[task_id] = task
			raise RuntimeError(f'task {task} already exists in being_done')
		self._being_done_ids.add(task_id)
		return task

	def add_done_task(self, task):
		"""
		:type task: Task
		"""
		if not isinstance(task, Task):
			raise TypeError(f'task should be of type Task but it is of type {type(task)}')
		if task.id not in self._being_done_ids:
			raise KeyError(f'task_id {task.id} does not exist in being_done_ids')
		if task.id in self._done:
			raise KeyError(f'task {task} already exists in done!')

		self.process(task=task)
		self._being_done_ids.remove(task.id)
		self._done[task.id] = task

	def process(self, task):
		raise NotImplementedError(f'this method should be implemented for class {self.__class__}')

	def produce_tasks(self, **kwargs):
		raise NotImplementedError(f'this method should be implemented for class {self.__class__}')

	def fill_to_do_list(self, num_tasks, echo=True, **kwargs):
		raise NotImplementedError(f'this method should be implemented for class {self.__class__}')

	def send_to_do(self, num_tasks=None, echo=True, **kwargs):
		self._processor.receive_to_do(
			project_name=self.name, num_tasks=num_tasks, echo=echo, process_done_tasks=True, **kwargs
		)

	def do(self, num_tasks=None, echo=True, disable_warnings=True, **kwargs):
		if not self._all_tasks_produced:
			self.produce_tasks(ignore_error=True, echo=echo)
		num_tasks = min(num_tasks, self.to_do_count + self.new_count)
		progress_bar = ProgressBar(total=num_tasks, echo=echo)

		num_done = 0
		to_do_count = self.to_do_count
		for i in range(to_do_count):
			progress_bar.show(amount=num_done, text=f'done: {num_done} / {to_do_count} (to-do)')
			task = self.pop_to_do()
			if disable_warnings:
				with warnings.catch_warnings():
					warnings.simplefilter('ignore')
					task.do(namespace=self.processor.namespace, worker_id='main')
			else:
				task.do(namespace=self.processor.namespace, worker_id='main')
			self.add_done_task(task=task)
			num_done += 1
		progress_bar.show(amount=num_done, text=f'done: {num_done} / {to_do_count} (to-do)')

		while num_tasks > num_done:
			progress_bar.show(amount=num_done, text=f'done: {num_done} / {num_tasks} (all tasks)')
			self.fill_to_do_list(num_tasks=1, echo=False, **kwargs)
			task = self.pop_to_do()
			if disable_warnings:
				with warnings.catch_warnings():
					warnings.simplefilter('ignore')
					task.do(namespace=self.processor.namespace, worker_id='main')
			else:
				task.do(namespace=self.processor.namespace, worker_id='main')
			self.add_done_task(task=task)
			num_done += 1
		progress_bar.show(amount=num_done, text=f'done: {num_done} / {num_tasks} (all tasks)')



SECONDS_IN_A_DAY = 60.0 * 60.0 * 24.0


def convert(delta, to_unit='timedelta'):
	u = to_unit[0].lower()
	if u == 't':
		return delta

	# microsecond
	elif to_unit[:2] == 'us' or to_unit[:3] == 'mic':
		return delta.days*SECONDS_IN_A_DAY*1e6 + delta.seconds*1e6 + delta.microseconds

	# millisecond
	elif to_unit[:2] == 'ms' or to_unit[:3] == 'mil':
		return delta.days*SECONDS_IN_A_DAY*1e3 + delta.seconds*1e3 + delta.microseconds / 1E3

	# second
	elif u == 's':
		return delta.days*SECONDS_IN_A_DAY + delta.seconds + delta.microseconds / 1E6

	# minute
	elif to_unit[:3] == 'min':
		return (delta.days*SECONDS_IN_A_DAY + delta.seconds + delta.microseconds / 1E6) / 60

	# hour
	elif u == 'h':
		return (delta.days*SECONDS_IN_A_DAY + delta.seconds + delta.microseconds / 1E6) / 3600

	# day
	elif u == 'd':
		return delta.days + delta.seconds/SECONDS_IN_A_DAY + delta.microseconds / 1E6 / SECONDS_IN_A_DAY

	else:
		raise ValueError(f'Cannot convert time delta to {to_unit}')



from pandas import concat, DataFrame
from ...exceptions import MissingArgumentError, AmbiguousArgumentsError
from ._split_data import split_data

from ._TrainingTestContainer import TrainingTestContainer
from ._ValidationContainer import ValidationContainer


def get_cross_validation(
		data, num_splits, holdout_ratio=None, holdout_count=None, sort_columns=None, id_columns=None,
		test_ratio=None, test_count=None, random_state=None,
		min_training_ratio=None, min_training_count=None
):
	"""
	:type 	data: DataFrame
	:type 	num_splits: int
	:type 	holdout_ratio: float
	:type 	holdout_count: int
	:type 	sort_columns: list[str]
	:type 	id_columns: list[str]
	:type 	test_ratio: float
	:type 	test_count: int
	:type 	random_state: int
	:type 	min_training_ratio: float
	:type 	min_training_count: int
	:rtype: ValidationContainer
	"""
	if id_columns is None and sort_columns is not None:
		id_columns = sort_columns
		time_series = True
	elif id_columns is not None and sort_columns is None:
		time_series = False
	elif id_columns is not None and sort_columns is not None:
		raise ValueError('Only one of sort_columns and id_columns can be None')
	else:
		time_series = False

	if test_ratio is None and test_count is None:
		test_ratio = holdout_ratio
		test_count = holdout_count
	elif test_ratio is not None and test_count is not None:
		raise AmbiguousArgumentsError('Only one of test_ratio and test_count should be provided!')

	if id_columns is None or id_columns == []:
		new_data = data.copy()
		id_columns = ['cross_validation_id']
		new_data['cross_validation_id'] = range(new_data.shape[0])
	else:
		if isinstance(id_columns, str):
			id_columns = [id_columns]
		new_data = data[id_columns].copy()

	new_data['cross_validation_index'] = range(new_data.shape[0])
	new_data['cross_validation_count'] = 1
	data_ids = new_data[id_columns + ['cross_validation_count']].groupby(id_columns).sum().reset_index(drop=False)
	total_count = data_ids['cross_validation_count'].sum()

	if holdout_ratio is not None and holdout_count is not None:
		raise AmbiguousArgumentsError('Only one of holdout_count and holdout_ratio should be provided!')
	elif holdout_ratio is not None:
		validation_holdout_counts = None
		validation_holdout_ratios = [1 - holdout_ratio, holdout_ratio]
	else:
		holdout_count = holdout_count or 0
		validation_holdout_counts = [total_count - holdout_count, holdout_count]
		validation_holdout_ratios = None

	# randomize for regular cross validation or sort for timeseries cross validation
	if time_series:
		data_ids = data_ids.sort_values(by=id_columns)
	else:
		data_ids = data_ids.sample(frac=1, random_state=random_state)

	validation_ids, holdout_ids = split_data(
		data=data_ids, count_column='cross_validation_count',
		counts=validation_holdout_counts, ratios=validation_holdout_ratios
	)

	def _get_specific_data(_id_data):
		if _id_data is None:
			return None
		return _id_data.merge(right=new_data, on=id_columns, how='left')  # .reset_index(drop=True)

	def _get_indices_from_specific_data(_specific_data):
		if _specific_data is None:
			return None
		return sorted(list(_specific_data['cross_validation_index']))

	validation_data = _get_specific_data(validation_ids)
	validation_indices = _get_indices_from_specific_data(validation_data)
	holdout_data = _get_specific_data(holdout_ids)
	holdout_indices = _get_indices_from_specific_data(holdout_data)

	if time_series:
		if num_splits + 1 > validation_ids.shape[0]:
			raise ValueError(f'maximum number of splits for this data is {validation_ids.shape[0] - 1}')
	else:
		if num_splits > validation_ids.shape[0]:
			raise ValueError(f'maximum number of splits for this data is {validation_ids.shape[0]}')

	if time_series:
		validation_ids = validation_ids.sort_values(by=id_columns)
	else:
		validation_ids = validation_ids.sample(frac=1, random_state=random_state)
	validation_ids = validation_ids  # .reset_index(drop=True)

	validation_total = validation_ids['cross_validation_count'].sum()

	if test_ratio is not None and test_count is not None:
		raise AmbiguousArgumentsError('Only one of test_count and test_ratio should be provided!')

	if num_splits == 0:
		fold_indices = []

	elif num_splits == 1:
		if test_ratio is not None:
			training_test_counts = None
			training_test_ratios = [1 - test_ratio, test_ratio]
		elif test_count is None:
			raise MissingArgumentError('Either test count or test ratio should be provided!')
		else:
			training_test_counts = [validation_total - test_count, test_count]
			training_test_ratios = None

		training_ids, test_ids = split_data(
			data=validation_ids, count_column='cross_validation_count',
			counts=training_test_counts, ratios=training_test_ratios
		)

		training_indices = _get_indices_from_specific_data(_get_specific_data(training_ids))
		test_indices = _get_indices_from_specific_data(_get_specific_data(test_ids))
		fold_indices = [{'training': training_indices, 'test': test_indices}]

	else:

		fold_indices = []
		if time_series:
			if min_training_count is None:
				if min_training_ratio is None:
					min_training_count = 0
				else:
					min_training_count = min_training_ratio * validation_ids.shape[0]

			first_count = max(min_training_count, validation_total / (num_splits + 1))
			other_counts = (validation_total - first_count) / num_splits
			counts = [first_count] + [other_counts] * num_splits
			splits = split_data(
				data=validation_ids, count_column='cross_validation_count',
				counts=counts, ratios=None
			)
			for i in range(num_splits):
				training_ids = concat(splits[:i + 1])
				test_ids = splits[i + 1]
				training_indices = _get_indices_from_specific_data(_get_specific_data(training_ids))
				test_indices = _get_indices_from_specific_data(_get_specific_data(test_ids))
				fold_indices.append({'training': training_indices, 'test': test_indices})

		else:

			counts = [validation_total / num_splits] * num_splits
			splits = split_data(
				data=validation_ids, count_column='cross_validation_count',
				counts=counts, ratios=None
			)
			for i in range(num_splits):
				training_ids = concat([splits[j] for j in range(num_splits) if j != i])
				test_ids = splits[i]
				training_indices = _get_indices_from_specific_data(_get_specific_data(training_ids))
				test_indices = _get_indices_from_specific_data(_get_specific_data(test_ids))
				fold_indices.append({'training': training_indices, 'test': test_indices})

	if fold_indices is not None:
		folds = [
			TrainingTestContainer(
				data=data, training_indices=fold['training'], test_indices=fold['test'],
				sort_columns=sort_columns
			)
			for fold in fold_indices
		]
	else:
		folds = []

	return ValidationContainer(
		data=data, validation_indices=validation_indices, holdout_indices=holdout_indices,
		folds=folds,
		sort_columns=sort_columns
	)



	return TrainingTestContainer(
		data=data, training_indices=validation.validation_indices, test_indices=validation.holdout_indices,
		sort_columns=sort_columns
	)




from ...validation import Validation
from .._DataSlice import TrainingTestSlice
from ._LearningProject import LearningProject
from ._LearningTask import LearningTask
from pandas import DataFrame


class CrossValidationProject(LearningProject):
	def __init__(
			self, name, y_column, problem_type, x_columns=None,
			time_unit='ms', evaluation_function=None, main_metric=None,
			lowest_is_best=None, best_score=None,
			scoreboard=None, processor=None
	):
		super().__init__(
			name=name, y_column=y_column, problem_type=problem_type, x_columns=x_columns,
			time_unit=time_unit, evaluation_function=evaluation_function, main_metric=main_metric,
			lowest_is_best=lowest_is_best, best_score=best_score, scoreboard=scoreboard, processor=processor
		)
		self._validation_holdout_slice_id = None
		self._full_data_slice_id = None

	def add_validation(
			self, data, validation,
			id_prefix='fold_',
			overwrite=False, random_state=None
	):
		"""

		:param data:
		:type  data: DataFrame

		:param validation:
		:type  validation: Validation

		:param processor:

		:param id_prefix:
		:type  id_prefix: str

		:type  overwrite: bool
		:type  random_state: int
		"""
		container = validation.split(data=data, random_state=random_state)
		data_id = self.name
		self.processor.add_data(data_id=data_id, data=container.data, overwrite=overwrite)

		for i, fold in enumerate(container.folds):
			training_test_slice_id = f'{self.name}_{id_prefix}{i + 1}'
			training_test_slice = TrainingTestSlice(
				data_id=data_id, training_indices=fold.training_indices, test_indices=fold.test_indices,
				columns=None
			)

			self.add_training_test_slice(
				training_test_slice_id=training_test_slice_id,
				training_test_slice=training_test_slice,
				data=None,
				overwrite=overwrite
			)

		if container.holdout_indices is not None:
			validation_holdout_slice = TrainingTestSlice(
				data_id=data_id, training_indices=container.validation_indices,
				test_indices=container.holdout_indices, columns=None
			)
			self._add_validation_holdout_slice(
				validation_holdout_slice_id=f'{self.name}_validation_holdout',
				validation_holdout_slice=validation_holdout_slice,
				data=None,
				overwrite=overwrite
			)

		# add full data
		full_data_slice = TrainingTestSlice(
			data_id=data_id, training_indices=None,
			test_indices=None,
			columns=None
		)
		self._add_full_data_slice(
			full_data_slice_id=f'{self.name}_full_data',
			full_data_slice=full_data_slice, data=None, overwrite=overwrite
		)

	def _add_validation_holdout_slice(
			self, validation_holdout_slice, data=None, validation_holdout_slice_id=None, overwrite=False
	):
		if validation_holdout_slice_id is None:
			validation_holdout_slice_id = f'{self.name}_validation_holdout'

		self._validation_holdout_slice_id = validation_holdout_slice_id

		self.add_training_test_slice(
			training_test_slice_id=validation_holdout_slice_id,
			training_test_slice=validation_holdout_slice,
			data=data,
			overwrite=overwrite,
			add_to_training_test_ids=False
		)

	def _add_full_data_slice(
			self, full_data_slice, data=None, full_data_slice_id=None, overwrite=False
	):
		if full_data_slice_id is None:
			full_data_slice_id = f'{self.name}_full_data'

		self._full_data_slice_id = full_data_slice_id

		self.add_training_test_slice(
			training_test_slice_id=full_data_slice_id,
			training_test_slice=full_data_slice,
			data=data,
			overwrite=overwrite,
			add_to_training_test_ids=False
		)

	def get_holdout_results(self):
		best_estimator = self.get_best_estimator()
		task = LearningTask(
			project_name=self.name, estimator_class=best_estimator['class'],
			estimator_name=best_estimator['name'],
			estimator_id=best_estimator['id'],
			estimator_arguments=best_estimator['arguments'],
			training_test_slice_id=self._validation_holdout_slice_id,
			y_column=self.y_column, x_columns=self.x_columns,
			evaluation_function=self.evaluation_function
		)
		task.do(namespace=self.processor.namespace, worker_id='main', return_predictions=True)
		return task

	def get_complete_model(self):
		best_estimator = self.get_best_estimator()
		task = LearningTask(
			project_name=self.name, estimator_class=best_estimator['class'],
			estimator_name=best_estimator['name'],
			estimator_id=best_estimator['id'],
			estimator_arguments=best_estimator['arguments'],
			training_test_slice_id=self._full_data_slice_id,
			y_column=self.y_column, x_columns=self.x_columns,
			evaluation_function=self.evaluation_function
		)
		task.do(namespace=self.processor.namespace, worker_id='main', return_predictions=True)
		return task



from multiprocess.managers import Namespace
from time import sleep

def worker(worker_id, namespace, to_do, doing, done, proceed, status):
	"""
	:type worker_id: int or str
	:type namespace: Namespace
	:type to_do: list[TrainingTestTask]
	:type doing: dict[int or str, Task]
	:type done: list[Task]
	:type proceed: dict[str, bool]
	:type status: dict[str, str]

	each item in the queue is tuple or list that has:
	estimator_id, data_id, estimator class, dictionary of kwargs, training, test

	"""
	status[worker_id] = 'started'
	if worker_id in proceed:
		error = ValueError(f'{worker_id} already exists in proceed')
		status[worker_id] = f'error: {error}'
		raise error
	else:
		proceed[worker_id] = True

	while proceed[worker_id]:
		try:
			task = to_do.pop(0)
			doing[worker_id] = task
			status[worker_id] = 'active'

		except IndexError:
			status[worker_id] = 'idle'
			continue

		try:
			task.do(namespace=namespace, worker_id=worker_id)

		except Exception as error:
			task.add_error(error=error)
		sleep(0.1)

		del doing[worker_id]
		done.append(task)

	status[worker_id] = 'ended'



from ..time import get_now


class WorkerReport:
	def __init__(self, worker_id):
		self._worker_id = worker_id
		self._start_time = get_now()
		self._end_time = None
		self._tasks = set()

	def add_task_id(self, task_id):
		self._tasks.add(task_id)

	def end(self):
		self._end_time = get_now()

	def __repr__(self):
		return f'Worker {self._worker_id} | {self._start_time}-{self._end_time} | {len(self._tasks)} tasks done.'

	@property
	def record(self):
		return {
			'worker_id': self._worker_id,
			'start_time': self._start_time,
			'end_time': self._end_time,
			'task_count': len(self._tasks)
		}


class SupervisorReport:
	def __init__(self, supervisor_id):
		self._supervisor_id = supervisor_id
		self._start_time = get_now()
		self._end_time = None
		self._logs = []

	def log(self, string):
		self._logs.append((get_now(), string))

	def __repr__(self):
		return f'Supervisor {self._supervisor_id} | {self._start_time}-{self._end_time} | {len(self._logs)} tasks done.'

	@property
	def records(self):
		return [
			{'time': log[0], 'event': log[1]}
			for log in self._logs
		]

	def end(self):
		self._end_time = get_now()


from pandas import DataFrame as PandasDF
from pyspark.sql.session import SparkSession
from pyspark.sql import DataFrame as SparkDF
from pyspark.sql.types import IntegerType, DateType, StringType
from pyspark.sql import functions as f
from bisect import bisect_right, bisect_left
from datetime import date as Date
from datetime import timedelta
from multiprocessing.pool import ThreadPool
from ...time.date import parse_date, find_date_and_format


class NearestFinder:
	"""
	sometimes we have a dataset with only a handful of values (e.g., dates) but we need to use it
	for any value within the range.
	this class creates a mapping from each date to the last date from a list of dates which is equal or before
	the date in question
	"""
	def __init__(self, direction='before', time_complexity='O(1)', parse_int_as_date_if_possible=True, n_jobs=8):
		"""

		:type time_complexity: str
		:param time_complexity: O(1) --> use dictionary, O(log(n)) --> use binary search (takes less memory)
		"""
		if direction.lower() not in ['before', 'after']:
			raise ValueError(f'direction "{direction}" is not supported.')
		self._direction = direction.lower()

		if time_complexity.lower() not in {'o(1)', 'o(log(n))'}:
			raise ValueError(f'Time complexity {time_complexity} is not supported!')

		if time_complexity == 'o(1)':
			self._function_type = 'map'
		else:
			self._function_type = 'binary_search'

		self._list = None
		self._bisect_function = None
		self._type = None
		self._represents = None
		self._date_format = None
		self._parse_int_as_date_if_possible = parse_int_as_date_if_possible
		self._column = None
		self._n_jobs = n_jobs
		self._mapping = None

	def __getstate__(self):
		attributes = [
			'_direction', '_mapping',
			'_list', '_type', '_represents', '_date_format', '_parse_int_as_date_if_possible', '_column', '_n_jobs'
		]
		return {
			name: getattr(self, name)
			for name in attributes
		}

	def __setstate__(self, state):
		for name, value in state.items():
			setattr(self, name, value)

		if self._function_type == 'map':
			self._create_map_function()
		else:
			self._create_binary_search_function()

	@staticmethod
	def _to_list(X, return_column=False):
		if isinstance(X, (SparkDF, PandasDF)):
			column = list(X.columns)[0]
			if len(X.columns) > 1:
				raise ValueError(f'DataFrame has {len(X.columns)} columns! Which one should I use?')

			if isinstance(X, SparkDF):
				X = X.toPandas()

			if isinstance(X, PandasDF):
				X = X[column]
		else:
			column = None

		if return_column:
			return list(X), column
		else:
			return list(X)

	@property
	def _first(self):
		return self._list[0]

	@property
	def _last(self):
		return self._list[-1]

	def fit(self, X):
		l, self._column = self._to_list(X=X, return_column=True)
		self._list = sorted(l)

		if type(self._first) != type(self._last):
			raise TypeError(f'different types are used in the list {type(self._first)}, {type(self._last)}')

		if self._function_type == 'map':
			self._create_map_function()
		else:
			self._create_binary_search_function()

		return self

	def _create_binary_search_function(self):
		if self._direction == 'before':
			def function(x):
				index = bisect_right(self._list, x) - 1
				index = max(0, index)
				return self._list[index]

		else:
			max_index = len(self._list) - 1
			def function(x):
				index = bisect_left(self._list, x)
				index = min(max_index, index)
				return self._list[index]

		self._bisect_function = function

	def _create_map_function(self):
		if self._bisect_function is None:
			self._create_binary_search_function()

		if isinstance(self._first, int):
			self._type = 'int'
		elif isinstance(self._first, Date):
			self._type = 'date'
		elif isinstance(self._first, str):
			self._type = 'str'

		date1, format1 = find_date_and_format(self._first)
		date2 = parse_date(self._last)
		if date1 is not None:
			self._represents = 'date'
			self._date_format = format1
		else:
			self._represents = self._type
			self._date_format = None

		if self._represents == 'date':
			with ThreadPool(self._n_jobs) as pool:
				values = pool.map(lambda n: date1 + timedelta(days=n), range((date2 - date1).days))

			if self._type == 'str' or self._type == 'int':
				with ThreadPool(self._n_jobs) as pool:
					values = pool.map(lambda x: x.strftime(self._date_format), values)

			if self._type == 'int':
				with ThreadPool(self._n_jobs) as pool:
					values = pool.map(lambda x: int(x), values)

		elif self._represents == 'int':
			values = list(range(self._list[-1] - self._list[0] + 1))

		else:
			values = None

		if values is not None:
			self._mapping = {x: self._bisect_function(x) for x in values}

			def function(x):
				return self._mapping[x]
		else:
			self._mapping = {x: x for x in self._list}

			def function(x):
				if x in self._mapping:
					return x
				else:
					y = self._bisect_function(x)
					self._mapping[x] = y
					return y

		self._mapping_function = function

	@property
	def mapping(self):
		"""
		:rtype: dict
		"""
		if self._mapping is None:
			self._create_map_function()
		return self._mapping

	def to_spark_df(self, spark=None, from_column='from', to_column='to'):
		"""
		:rtype: SparkDF
		"""
		if spark is None:
			spark = SparkSession.builder.getOrCreate()

		return spark.createDataFrame(
			[(key, value) for key, value in self.mapping.items()],
			[from_column, to_column]
		)

	def to_pandas_df(self, from_column='from', to_column='to'):
		"""
		:rtype: PandasDF
		"""
		return PandasDF({from_column: self.mapping.keys(), to_column: self.mapping.values()})

	def map(self, x):
		if self._function_type == 'dict':
			return self._mapping_function(x)
		else:
			return self._bisect_function(x)

	def predict(self, X):
		if isinstance(X, SparkDF):
			if self._column not in X.columns:
				raise ValueError(f'Column {self._column} does not exist in X.')

			if self._type == 'date':
				the_type = DateType()
			elif self._type == 'int':
				the_type = IntegerType()
			else:
				the_type = StringType()

			@f.udf(the_type)
			def _f(x):
				return self.map(x=x)

			return X.withColumn('prediction', _f(self._column))

		elif isinstance(X, PandasDF):
			if self._column not in X.columns:
				raise ValueError(f'Column {self._column} does not exist in X.')

			return X[self._column].apply(self.map)

		elif isinstance(X, list):
			with ThreadPool(self._n_jobs) as pool:
				result = pool.map(self.map, X)
			return result

		else:
			raise TypeError(f'type {type(X)} is not supported!')



from ..ProcessController import ProcessController
from ._validate import validate


class

class ValidationController(ProcessController):
	def validate(self, validation_container,
		estimators, evaluation_type, x_columns=None, y_column=None, aggregate=False,
		echo=0
	):
		"""
		performs the _evaluate function of ValidationContainer in a multiprocessing manner
		:param validation_container:
		:type  validation_container: ValidationContainer
		:param controller:
		:type  controller: ProcessController
		:param estimators: the estimators to be evaluated
		:type  estimators: list or dict
		:param evaluation_type: regression or classification
		:type  evaluation_type: str
		:type  x_columns: list[str]
		:type  y_column: str
		:param aggregate: if the results should be aggregated per estimator
		:type  echo: int or bool
		:rtype: DataFrame
		"""
		return validate(
			validation_container=validation_container, controller=self, estimators=estimators,
			evaluation_type=evaluation_type, x_columns=x_columns, y_column=y_column, aggregate=aggregate,
			echo=echo
		)



from pandas import DataFrame

def join_and_keep_order(left, right, remove_duplicates, keep='first', **kwargs):
	"""
	:type left: DataFrame
	:type right: DataFrame
	:rtype: DataFrame
	"""
	left = left.copy()
	right = right.copy()
	left['_left_id'] = range(left.shape[0])
	right['_right_id'] = range(right.shape[0])
	result = left.merge(right=right, **kwargs)
	result.sort_values(axis='index', by=['_left_id', '_right_id'], inplace=True)
	if remove_duplicates: result = result[(~result['_left_id'].duplicated(keep=keep)) & (~result['_right_id'].duplicated(keep=keep))]
	return result.drop(columns=['_left_id', '_right_id'])


def join_wisely(left, right, remove_duplicates=True, echo=False, **kwargs):
	"""
	joins two dataframes and returns a dictionary with 3 members: left_only, right_only, and both (the results of the two joins)
	:type left: DataFrame
	:type right: DataFrame
	:type kwargs: dict
	:rtype: dict of DataFrames
	"""
	left=left.copy()
	right=right.copy()

	left['_left_id'] = range(left.shape[0])
	right['_right_id'] = range(right.shape[0])

	both_data = left.merge(right=right, how='inner', **kwargs)
	if remove_duplicates:
		both_data.sort_values(axis=0, by=['_left_id', '_right_id'], inplace=True)
		left_id_duplicated = both_data._left_id.duplicated(keep='first')
		right_id_duplicated = both_data._right_id.duplicated(keep='first')
		both_data = both_data[~left_id_duplicated & ~right_id_duplicated]
	left_only_data = left[~left['_left_id'].isin(both_data['_left_id'])].copy()
	right_only_data = right[~right['_right_id'].isin(both_data['_right_id'])].copy()

	both_data.drop(labels=['_left_id', '_right_id'], axis=1, inplace=True)
	left_only_data.drop(labels='_left_id', axis=1, inplace=True)
	right_only_data.drop(labels='_right_id', axis=1, inplace=True)
	if echo:
		print(f'left:{left.shape}, right:{right.shape}\nboth:{both_data.shape}, left_only:{left_only_data.shape}, right_only:{right_only_data.shape}')

	return {'both':both_data, 'left_only':left_only_data, 'right_only':right_only_data}




from datetime import date


def get_today():
	"""
	:rtype: date
	"""
	return date.today()


def get_today_str():
	"""
	:rtype:  str
	"""
	return str(get_today())


def get_date(year=None, month=None, day=None):
	year = year or get_today().year
	month = month or get_today().month
	day = day or get_today().day
	return date(year, month, day)



from pandas import DataFrame
from ...hash import hash_object
from ...exceptions import MissingColumnsError, MissingRowsError
from ...collections import OrderedSet
from ._get_rows import get_rows, get_columns


DATA_STATE_ATTRIBUTES = ['_columns', '_rows', '_original_data']


class DataSlice:
	def __init__(self, data, columns=None, rows=None):
		"""
		:type data: DataFrame or DataSlice
		"""
		self._parent = data
		if isinstance(self.parent, self.__class__):
			self._original_data = self.parent.original_data
		else:
			self._original_data = data

		if columns is None:
			self._columns = get_columns(self.parent)
		else:
			columns = OrderedSet(columns)
			parent_columns = get_columns(self.parent)
			if columns <= parent_columns:
				self._columns = columns
			else:
				raise MissingColumnsError(f'missing columns: {columns - parent_columns}')

		if rows is None:
			self._rows = get_rows(self.parent)
		else:
			rows = OrderedSet(rows)
			parent_rows = get_rows(self.parent)
			if rows <= parent_rows:
				self._rows = rows
			else:
				raise MissingRowsError(f'missing rows: {rows - parent_rows}')

	@property
	def parent(self):
		"""
		:rtype: DataFrame or DataSlice
		"""
		return self._parent

	@property
	def original_data(self):
		return self._original_data

	def __getstate__(self):
		return {key: getattr(self, key) for key in DATA_STATE_ATTRIBUTES}

	def __setstate__(self, state):
		for key, value in state.items():
			setattr(self, key, value)

	def get_data_hash(self):
		return hash_object(self.data)

	def __hashkey__(self):
		return self.original_data, self.columns, self.rows

	@property
	def data(self):
		"""
		:rtype: DataFrame
		"""
		return self.original_data[self.columns].iloc[self.rows]

	@property
	def columns(self):
		"""
		:rtype: list[str]
		"""
		return self._columns

	@property
	def rows(self):
		"""
		:rtype: list[int]
		"""
		return self._rows

	@property
	def num_columns(self):
		return self.shape[1]

	@property
	def num_rows(self):
		return self.shape[0]

	@property
	def shape(self):
		return self.data.shape

	def __repr__(self):
		return f'<Data: {self.num_rows}x{self.num_columns}>'



def get_function_arguments(function):
	"""
	:param callable function: a function
	:rtype: list[str]
	"""
	function_arguments = list(function.__code__.co_varnames)[:function.__code__.co_argcount]
	return function_arguments



from atlantis.ds.synthetic_data import create_data
from atlantis.ds.validation import CrossValidation
from atlantis.ds.validation import EstimatorRepository
from atlantis.ds.parallel_computing import Processor

from sklearn.linear_model import LinearRegression, Lasso
from sklearn.ensemble import RandomForestRegressor

repository = EstimatorRepository()
repository.append(RandomForestRegressor, {'n_estimators': 100, 'n_jobs': 1, 'max_depth': [6, 12, 24]})
repository.append(Lasso, {'alpha': [0, 0.1, 1, 10]})
repository.append(Lasso, {'alpha': [1, 10, 20]})
repository.append(LinearRegression)

data = create_data(num_rows=10000, num_x_columns=100, noise=2)
cv = CrossValidation(num_splits=5)

processor = Processor()
processor.add_workers(num_workers=8)

project = processor.create_cross_validation_project(name='example', y_column='y', problem_type='regression')
project.add_estimator_repository(repository=repository)
project.add_validation(data=data, validation=cv, random_state=42)
display(project)

project.send_to_do(num_tasks=10)

processor.show_progress()



from pandas import DataFrame

from ...ds.validation import ValidationContainer
from ...ds.validation._train_and_test import train_and_test
from .. import ProcessController
from ...time.progress import ProgressBar

from time import sleep


def validate(
		validation_container, controller,
		estimators, evaluation_type, x_columns=None, y_column=None, aggregate=False,
		echo=0
):
	"""
	performs the _evaluate function of ValidationContainer in a multiprocessing manner
	:param validation_container:
	:type  validation_container: ValidationContainer
	:param controller:
	:type  controller: ProcessController
	:param estimators: the estimators to be evaluated
	:type  estimators: list or dict
	:param evaluation_type: regression or classification
	:type  evaluation_type: str
	:type  x_columns: list[str]
	:type  y_column: str
	:param aggregate: if the results should be aggregated per estimator
	:type  echo: int or bool
	:rtype: DataFrame
	"""
	task_ids = set()
	for dictionary in validation_container._generate_evaluation(
		estimators=estimators, evaluation_type=evaluation_type, x_columns=x_columns, y_column=y_column
	):
		estimator = dictionary['kwargs']['estimator']
		try:
			cpu_count = estimator.n_jobs
		except AttributeError:
			cpu_count = 1

		if cpu_count == -1:
			cpu_count = controller.system_cpu_count

		task_id = controller.add_task(
			function=train_and_test,
			kwargs=dictionary['kwargs'],
			cpu_count=cpu_count
		)
		task_ids.add(task_id)

	total_tasks = len(task_ids)
	pb = ProgressBar(total=total_tasks + 1, echo=echo)
	total_processed = 0
	records = []
	while len(task_ids) > 0:
		pb.show(amount=total_processed, text='evaluating folds and estimators in parallel!')
		processed = controller.process_done_queue(echo=0)
		for task_id in processed:
			if task_id in task_ids:
				record = controller.processed[task_id].result
				total_processed += 1
				task_ids.remove(task_id)
				records.append(record)

		sleep(0.5)

	pb.show(amount=total_processed, text=f'aggregating results!')
	df = DataFrame.from_records(records)
	if aggregate:
		df = df.drop(columns='fold').groupby(
			['estimator_id', 'estimator_name']
		).agg(['mean', 'median', 'min', 'max', 'std'])
	pb.show(amount=total_processed + 1, text=f'evaluation complete!')

	return df



from pandas import DataFrame
from numpy import isnan


class Normalizer:
	def __init__(self, ignore_columns=None):
		self._columns = None
		self._means = None
		self._standard_deviations = None

		if isinstance(ignore_columns, str):
			ignore_columns = [ignore_columns]
		self._ignore_columns = ignore_columns or []

	def __getstate__(self):
		return {
			'_columns': self._columns, '_means': self._means, '_standard_deviations': self._standard_deviations,
			'_ignore_columns': self._ignore_columns,
		}

	def __setstate__(self, state):
		for key, value in state.items():
			setattr(self, key, value)

	def fit(self, X, y=None):
		"""
		:type X: DataFrame
		"""
		if self._means is not None or self._standard_deviations is not None:
			raise RuntimeError('normalizer is already fit!')

		self._means = {
			column: X[column].mean(skipna=True)
			for column in X.columns if column not in self._ignore_columns
		}

		self._standard_deviations = {
			column: X[column].std(skipna=True)
			for column in X.columns if column not in self._ignore_columns
		}

		nan_values = []
		for column, m in self._means.items():
			if isnan(m) or isnan(self._standard_deviations[column]):
				nan_values.append(str(column))
		if len(nan_values) > 0:
			raise ValueError(f'{", ".join(nan_values)} produce nan values!')

	def transform(self, X):
		"""
		:type X: DataFrame
		:rtype: DataFrame
		"""
		if self._means is None or self._standard_deviations is None:
			raise RuntimeError('normalizer is not fit!')

		result = X.copy()

		for column in self._means.keys():
			if self._standard_deviations[column] != 0:
				result[column] = (result[column] - self._means[column]) / self._standard_deviations[column]
			else:
				result[column] = (result[column] - self._means[column])

		return result

	def fit_transform(self, X, y=None):
		self.fit(X=X, y=y)
		return self.transform(X=X)

	def untransform(self, X):
		"""
		:type X: DataFrame
		:rtype: DataFrame
		"""
		if self._means is None or self._standard_deviations is None:
			raise RuntimeError('normalizer is not fit!')

		result = X.copy()

		for column in self._means.keys():
			if self._standard_deviations[column] != 0:
				result[column] = result[column] * self._standard_deviations[column] + self._means[column]
			else:
				result[column] = result[column] + self._means[column]

		return result

	@property
	def parameters(self):
		"""
		:rtype: DataFrame
		"""
		return DataFrame.from_records([
			{'column': column, 'mean': self._means[column], 'sd': self._standard_deviations[column]}
			for column in self._means.keys()
		])

	@classmethod
	def get_normalizer_and_normalized(cls, data, **kwargs):
		"""
		:rtype: tuple[Normalizer, DataFrame]
		"""
		normalizer = cls(**kwargs)
		normalized = normalizer.fit_transform(X=data)
		return normalizer, normalized


def get_normalizer_and_normalized(data, **kwargs):
	"""
	:rtype: tuple[Normalizer, DataFrame]
	"""
	return Normalizer.get_normalizer_and_normalized(data=data, **kwargs)



from .DataSlice import DataSlice
from pandas import DataFrame
from ...collections import OrderedSet


def get_and(x, y):
	if x is None:
		return y
	elif y is None:
		return x
	else:
		return x & y


def get_or(x, y):
	if x is None:
		return y
	elif y is None:
		return x
	else:
		return x | y


class SliceTemplate:
	def __init__(self, columns, rows):
		if columns is None:
			self._columns = None
		else:
			self._columns = OrderedSet(columns)

		if rows is None:
			self._rows = None
		else:
			self._rows = OrderedSet(rows)

	def get_slice(self, data):
		"""
		:type data: DataFrame
		:rtype: DataSlice
		"""
		return DataSlice(data=data, columns=self._columns, rows=self._rows)

	def __and__(self, other):
		return self.__class__(columns=get_and(self._columns, other._columns), rows=get_and(self._rows, other._rows))




from pandas import concat, DataFrame

# shuffles data and adds a column with new index
def shuffle(data, col = None):
	result = data.sample(frac = 1).reset_index(drop = True)
	if col is not None:
		result[col] = result.index

	return result


# sample the data but make sure to have a balanced number of samples per each group
# the groups are defined by columns
def sample_per_group(data, group_by, ratio=None, n=None):
	"""

	:type data: DataFrame
	:type group_by: list of str
	:type ratio: float
	:type num_rows: int
	:return:
	"""

	# group the data
	data = data.copy()
	"""
	:type data: DataFrame
	"""
	data['__order1'] = data.index
	grouped = data.groupby(by = group_by)

	training_sets = []
	test_sets = []

	for name, group in grouped:
		if n is None:
			num_rows = round(group.shape[0] * ratio)
		else:
			num_rows = round(min(n, group.shape[0]))
		shuffled = shuffle(data=group).reset_index(drop=True)
		new_training_set = shuffled.iloc[:num_rows, ]
		if num_rows>group.shape[0]:
			# if all of the data ends up in training set, test set should be empty instead of an error
			new_test_set = shuffled[shuffled.index!=shuffled.index]
		else:
			new_test_set = shuffled.iloc[(num_rows + 1):, ]
		training_sets.append(new_training_set)
		test_sets.append(new_test_set)

	training = concat(training_sets)
	"""
	:type training: DataFrame
	"""
	test = concat(test_sets)
	"""
	:type test: DataFrame
	"""

	training.reset_index(drop=True, inplace=True)
	test.reset_index(drop=True, inplace=True)

	training.index = training['__order1'].values
	test.index = test['__order1'].values
	training.sort_index(inplace=True)
	test.sort_index(inplace=True)

	training.drop(axis=1, labels = '__order1', inplace=True)
	test.drop(axis=1, labels='__order1', inplace=True)


	return training, test





from .Normalizer import Normalizer
from .Normalizer import get_normalizer_and_normalized
from .Polynomial import Polynomial
from .OneHotEncoder import OneHotEncoder
from .OneHotEncoder import get_one_hot_encoder_and_encoded
from .Scaler import Scaler
from .PCA import PCA
from .disassemble import disassemble



from sklearn.linear_model import LinearRegression
from sklearn.ensemble import RandomForestClassifier
from pandas import DataFrame, concat
from random import randint, random, choice
# from func_timeout import func_timeout, FunctionTimedOut # ToDo is this necessary?
# import matplotlib.pyplot as plt # ToDo is this necessary?
# from numpy import where # ToDo is this necessary?

from ...hash import hash_object
from ...functions import get_function_arguments
from ...ds.preprocessing import Polynomial, Normalizer

from .create_arguments import create_arguments
from .Measurement import Measurement
from ..time import get_elapsed
from ..time import get_now
from ..progress import ProgressBar


# Estimator gets a single argument function and estimates the time it takes to run the function based on the argument
# the function should accept an int larger than 0
class Estimator:
	def __init__(self, function, args=None, unit='s', polynomial_degree=2, timeout=20):
		self._function = function
		self._function_arguments = get_function_arguments(function=function)
		self._unit = unit
		self._measurements = {}
		self._polynomial_degree = polynomial_degree
		self._model = None
		self._normalizer = None
		self._error_model = None
		self._max_x = None
		self._args = args
		self._timeout = timeout
		self._num_errors = 0
		self._num_regular_runs = 0
		self._num_timeouts = 0
		self._x_data_columns = {}
		self._error_x_data_columns = {}

	@staticmethod
	def get_key(**kwargs):
		return hash_object(kwargs)

	def check_arguments(self, kwargs, method_name):
		unknown_arguments = [key for key in kwargs.keys() if key not in self._function_arguments]
		missing_arguments = [key for key in self._function_arguments if key not in kwargs]

		if len(missing_arguments) == 1:
			return f'{method_name}() is missing the argument "{missing_arguments[0]}"'
		elif len(missing_arguments) > 1:
			arguments_string = '", "'.join(missing_arguments)
			return f'{method_name}() is missing arguments "{arguments_string}"'

		if len(unknown_arguments) == 0:
			return False
		elif len(unknown_arguments) == 1:
			return f'{method_name}() got an unexpected argument "{unknown_arguments[0]}"'
		else:
			arguments_string = '", "'.join(unknown_arguments)
			return f'{method_name}() got unexpected arguments "{arguments_string}"'

	def get_arguments(self, arguments, **kwargs):
		if arguments is None and len(kwargs) == 0:
			raise ValueError('either arguments should be provided or kwargs!')
		elif arguments is not None and len(kwargs) > 0:
			raise ValueError('only one of arguments and kwargs should be provided!')
		elif arguments is None:
			arguments = kwargs
		return arguments

	def measure(self, timeout=None, arguments=None, **kwargs):
		"""
		:type timeout: int or float
		:type arguments: NoneType or dict
		:rtype: Measurement
		"""

		kwargs = self.get_arguments(arguments=arguments, **kwargs)

		if self.check_arguments(kwargs=kwargs, method_name='measure'):
			raise TypeError(self.check_arguments(kwargs=kwargs, method_name='measure'))

		key = self.get_key(**kwargs)
		if key in self._measurements:
			return self._measurements[key]
		else:
			start_time = get_now()
			if not timeout:
				try:
					result = self._function(**kwargs)
					timeout_error = False
					other_error = False
					self._num_regular_runs += 1

				except Exception as e:
					result = None
					timeout_error = False
					other_error = True
					self._num_errors += 1

			else:
				def run_function():
					return self._function(**kwargs)

				try:
					result = func_timeout(timeout, run_function)
					timeout_error = False
					other_error = False
					self._num_regular_runs += 1

				except FunctionTimedOut:
					result = None
					timeout_error = True
					other_error = False
					self._num_timeouts += 1

				except Exception as e:
					result = None
					timeout_error = False
					other_error = True
					self._num_errors += 1

			elapsed = get_elapsed(start=start_time, unit=self._unit)
			measurement = Measurement(
				x=kwargs, result=result, elapsed=elapsed, timeout_error=timeout_error, other_error=other_error
			)
			self._measurements[key] = measurement
			self._model = None
			self._normalizer = None
			self._error_model = None
			if self._max_x is None:
				self._max_x = kwargs
			else:
				self._max_x = {key: max(value, kwargs[key]) for key, value in self._max_x.items()}
			return measurement

	@property
	def data(self):
		"""
		:rtype: DataFrame
		"""
		return DataFrame.from_records(
			[measurement.dictionary for measurement in self.measurements]
		)

	@property
	def measurements(self):
		"""
		:rtype: list[Measurement]
		"""
		measurements = sorted(list(self._measurements.values()))
		# set the weights
		min_elapsed = measurements[0].elapsed_time
		for measurement in measurements:
			if min_elapsed > 0:
				measurement._weight = 1 + (measurement.elapsed_time / min_elapsed) ** 0.5
			else:
				measurement._weight = 1
		return measurements

	@property
	def num_measurements(self):
		"""
		:rtype: int
		"""
		return len(self._measurements)

	@property
	def num_errors(self):
		return self._num_errors

	@property
	def num_regular_runs(self):
		return self._num_regular_runs

	@property
	def num_timeouts(self):
		return self._num_timeouts

	@property
	def num_runs(self):
		return self.num_errors + self.num_regular_runs

	def get_x_data(self, x, degree=None):
		"""
		:type x: DataFrame or dict or list
		:type degree: NoneType or int
		:rtype: DataFrame
		"""
		if isinstance(x, dict):
			if all([isinstance(value, (list, tuple)) for value in x.values()]):
				data = DataFrame(x)
			else:
				data = DataFrame.from_records([x])

		elif isinstance(x, list) and all([isinstance(element, dict) for element in x]):
			data = DataFrame.from_records(x)

		elif isinstance(x, list):
			data = DataFrame({'x': x})

		else:
			data = x

		degree = degree or self._polynomial_degree

		if degree > 1:
			polynomial = Polynomial(degree=degree)
			result = polynomial.fit_transform(data=data)

		else:
			result = data.copy()

		if degree not in self._x_data_columns:
			self._x_data_columns[degree] = list(result.columns)
		return result[self._x_data_columns[degree]]

	def get_error_x_data(self, x):
		data = self.get_x_data(x=x, degree=1)

		for column in self._function_arguments:
			data[f'{column}_is_zero'] = data[column] == 0
			data[f'{column}_is_positive'] = 1 * (data[column] > 0)
			data[f'{column}_rounded'] = data[column].round()
			data[f'{column}_is_int'] = data[f'{column}_rounded'] == data[column]

		if 1 not in self._error_x_data_columns:
			self._error_x_data_columns[1] = list(data.columns)
		return data[self._error_x_data_columns[1]]

	@property
	def training_x_y_weights(self):
		"""
		:rtype: tuple[DataFrame, list, list]
		"""
		records = []
		y = []
		weights = []
		for measurement in self.measurements:
			if not measurement.timeout_error and not measurement.other_error:
				records.append(measurement.x)
				y.append(measurement.elapsed_time)
				weights.append(measurement.elapsed_time ** 2)
		x = self.get_x_data(DataFrame.from_records(records))
		return x, y, weights

	@property
	def error_training_x_y_weights(self):
		"""
		:rtype: tuple[DataFrame, list]
		"""
		records = []
		y = []
		weights = []
		error_count = self.num_errors
		other_count = len(self._measurements) - error_count

		for measurement in self._measurements.values():
			if not measurement.timeout_error:
				records.append(measurement.x)
				if measurement.other_error:
					y.append(1)
					weights.append(other_count)
				else:
					y.append(0)
					weights.append(error_count)
		x = self.get_error_x_data(DataFrame.from_records(records))
		return x, y, weights

	@property
	def model(self):
		"""
		:rtype: LinearRegression
		"""
		if self._model is None:
			x, y, weights = self.training_x_y_weights
			self._model = LinearRegression()
			normalized_x = self.normalizer.transform(X=x)
			self._model.fit(normalized_x, y, sample_weight=weights)
		return self._model

	@property
	def normalizer(self):
		"""
		:type Normalizer
		"""
		if self._normalizer is None:
			x, y, weights = self.training_x_y_weights
			self._normalizer = Normalizer()
			self._normalizer.fit(X=x)
		return self._normalizer

	@property
	def error_model(self):
		"""
		:rtype: RandomForestClassifier
		"""
		if self._error_model is None:
			x, y, weights = self.error_training_x_y_weights
			self._error_model = RandomForestClassifier()
			self._error_model.fit(x, y, sample_weight=weights)
		return self._error_model

	def is_in_measurements(self, **kwargs):
		return self.get_key(**kwargs) in self._measurements

	def predict(self, **kwargs):
		error_probability = self.predict_error_probability(**kwargs)
		time_prediction = self.predict_time(**kwargs)
		return {'error_probability': error_probability, 'time_prediction': time_prediction}

	def predict_data(self, data):
		"""
		:type data: DataFrame
		:rtype: DataFrame
		"""
		error_probability = list(1 - self.error_model.predict_proba(self.get_error_x_data(data))[:, 0])
		error = self.error_model.predict(self.get_error_x_data(data))
		normalized = self.normalizer.transform(self.get_x_data(data))
		elapsed = self.model.predict(normalized)
		return DataFrame({'error_probability': error_probability, 'error': error, 'time': elapsed})

	def predict_time(self, **kwargs):
		"""
		:rtype: list or float or int
		"""
		if self.check_arguments(kwargs=kwargs, method_name='predict'):
			raise TypeError(self.check_arguments(kwargs=kwargs, method_name='predict'))

		key = self.get_key(**kwargs)
		if key in self._measurements:
			return self._measurements[key].elapsed_time

		x_data = self.get_x_data(x=kwargs)
		normalized = self.normalizer.transform(x_data)
		predictions = self.model.predict(normalized)
		if x_data.shape[0] > 1:
			return list(predictions)
		else:
			return list(predictions)[0]

	def predict_error(self, **kwargs):
		"""
		:rtype: list or float or int
		"""
		if self.check_arguments(kwargs=kwargs, method_name='predict_error'):
			raise TypeError(self.check_arguments(kwargs=kwargs, method_name='predict_error'))

		key = self.get_key(**kwargs)
		if key in self._measurements:
			return 1 * self._measurements[key].other_error

		x_data = self.get_error_x_data(x=kwargs)
		predictions = self.error_model.predict(x_data)
		if x_data.shape[0] > 1:
			return list(predictions)
		else:
			return list(predictions)[0]

	def predict_error_probability(self, **kwargs):
		"""
		:rtype: list or float or int
		"""
		if self.check_arguments(kwargs=kwargs, method_name='predict_error'):
			raise TypeError(self.check_arguments(kwargs=kwargs, method_name='predict_error'))

		key = self.get_key(**kwargs)
		if key in self._measurements:
			return 1 * self._measurements[key].other_error

		x_data = self.get_error_x_data(x=kwargs)
		predictions = 1 - self.error_model.predict_proba(x_data)[:, 0]
		if x_data.shape[0] > 1:
			return list(predictions)
		else:
			return list(predictions)[0]

	@staticmethod
	def choose_value(values):
		if isinstance(values, list):
			return choice(values)
		elif isinstance(values, tuple) and len(values) == 2:
			if isinstance(values[0], int) and isinstance(values[1], int):
				return randint(values[0], values[1])
			else:
				return min(values) + random() * (max(values) - min(values))
		else:
			return values

	def auto_explore(self, timeout=None, max_order=5):
		timeout = timeout or self._timeout

		if self.num_measurements < 3:
			for i in range(3):
				self.measure(timeout=timeout, arguments={key: i for key in self._function_arguments})

		for main_key in self._function_arguments:
			for order_of_magnitude in range(max_order):
				argument_range = (-1 * 10 ** order_of_magnitude, 10 ** order_of_magnitude)
				print(main_key, argument_range)
				arguments = {key: argument_range if key == main_key else 1 for key in self._function_arguments}
				self.explore(
					timeout=timeout,
					num_rounds=10,
					arguments=arguments
				)

		for order_of_magnitude in range(max_order):
			argument_range = (-1 * 10 ** order_of_magnitude, 10 ** order_of_magnitude)
			print(argument_range)
			self.explore(
				timeout=timeout,
				num_rounds=10,
				arguments={key: argument_range for key in self._function_arguments}
			)

	def explore(self, timeout, num_rounds=1000, arguments=None, random_state=42, **kwargs):
		start_time = get_now()
		arguments = self.get_arguments(arguments=arguments, **kwargs)

		progress_bar = ProgressBar(total=timeout)
		elapsed = get_elapsed(start=start_time, unit=self._unit)
		progress_bar.show(amount=elapsed, total=timeout)

		candidates = create_arguments(num_rounds=num_rounds, arguments=arguments)

		# remove candidates that have already been tried
		tried_candidates = self.data.drop(columns=['time', 'result', 'timeout_error', 'other_error'])
		candidates = concat([candidates, tried_candidates, tried_candidates])
		candidates.drop_duplicates(keep=False, inplace=True)

		candidates['_id_'] = range(candidates.shape[0])
		candidates = candidates.sample(frac=1)

		elapsed = get_elapsed(start=start_time, unit=self._unit)
		num_points = num_errors = num_timeouts = max_elapsed = 0
		num_candidates = candidates.shape[0]

		while timeout > elapsed and num_candidates > 0:
			progress_bar.show(
				amount=elapsed,
				total=timeout,
				text=f'n:{num_points}, e:{num_errors}, o:{num_timeouts}, t:{round(max_elapsed, 2)}, c:{num_candidates}'
			)
			candidates.reset_index(drop=True, inplace=True)
			predictions = self.predict_data(data=candidates.drop(columns='_id_'))

			predictions.reset_index(drop=True, inplace=True)
			predictions['_id_'] = candidates['_id_']

			# if we are out of time, stop
			elapsed = get_elapsed(start=start_time, unit=self._unit)
			remaining_time = timeout - elapsed
			if remaining_time < 0:
				break

			next_candidates = candidates.copy()

			# if we have had too many errors, try to not go after errors
			if self.num_errors > self._num_regular_runs and predictions['error'].sum() < predictions.shape[0]:
				next_candidates = candidates[predictions['error'] == 0]
				predictions = predictions[predictions['error'] == 0]

			# only try the ones that are reasonable timewise
			if (predictions['time'] < remaining_time).sum() > 0:
				next_candidates = next_candidates[predictions['time'] < remaining_time]
				predictions = predictions[predictions['time'] < remaining_time]
			else:
				# there will be no candidate, let's at least try our best with the shortest elapsed time
				next_candidates = next_candidates[predictions['time'] == predictions['time'].min()].head(1)
				predictions = predictions[predictions['time'] == predictions['time'].min()].head(1)

			# stop if there are no candidates
			if next_candidates.shape[0] == 0:
				break

			# choose longest time:
			# best_candidate = next_candidates[predictions['time'] == predictions['time'].max()].head(1)

			# choose randomly
			best_candidate = next_candidates.sample(n=1, random_state=random_state)
			x = best_candidate.to_dict('records')[0]
			best_candidate_id = x['_id_']
			del x['_id_']
			measurement = self.measure(timeout=timeout, **x)

			# remove the candidate that was measured
			candidates = candidates[candidates['_id_'] != best_candidate_id]
			num_candidates = candidates.shape[0]

			if measurement.timeout_error:
				num_timeouts += 1
			elif measurement.other_error:
				num_errors += 1
			else:
				num_points += 1
				max_elapsed = max(max_elapsed, measurement.elapsed_time)
			elapsed = get_elapsed(start=start_time, unit=self._unit)

		progress_bar.show(
			amount=timeout + 1,
			total=timeout,
			text=f'n:{num_points}, e:{num_errors}, o:{num_timeouts}, t:{round(max_elapsed, 2)}, c:{num_candidates}'
		)

	def plot(self, x, ignore_errors=True, marker='.', linestyle='', ms=12):
		data = self.data.copy()
		data['error'] = where(data.timeout_error, 'timeout', where(data.other_error, 'error', 'normal'))
		data.drop(columns=['timeout_error', 'other_error'], inplace=True)

		if ignore_errors:
			fig, ax = plt.subplots()
			ax.plot(data[x], data['time'], marker=marker, linestyle=linestyle, ms=ms)

		else:
			groups = data.groupby('error')
			fig, ax = plt.subplots()
			ax.margins(0.05)  # Optional, just adds 5% padding to the autoscaling
			for name, group in groups:
				ax.plot(group[x], group['time'], marker=marker, linestyle=linestyle, ms=ms, label=name)
			ax.legend()
		plt.show()
		return fig, ax



class TimeMeasurement:
	def __init__(self, duration, unit, count=1):
		self._duration = duration
		self._unit = unit
		self._count = count

	def __hashkey__(self):
		return (self.__class__.__name__, self.unit, self.duration, self.count)

	def copy(self):
		return self.__class__(duration=self._duration, unit=self._unit, count=self._count)

	@property
	def unit(self):
		return self._unit

	@property
	def count(self):
		return self._count

	@property
	def duration(self):
		return self._duration

	@property
	def mean_duration(self):
		return self.duration / self.count

	def __add__(self, other):
		if self.unit != other.unit:
			raise TypeError('cannot add two measurements of different type!')
		return self.__class__(
			duration=self.duration + other.duration,
			unit=self.unit, count=self.count + other.count
		)




from pandas import DataFrame
from ...exceptions import MissingArgumentError
from ._Evaluation import Evaluation
from ..evaluation import evaluate_regression, evaluate_classification
from ._DataContainer import get_display_function, DataContainer
from ...time.progress import ProgressBar


class TrainingTestContainer(DataContainer):
	def __init__(
			self, data, training_indices, test_indices, x_columns=None, y_column=None,
			sort_columns=None
	):
		super().__init__(data=data, x_columns=x_columns, y_column=y_column, sort_columns=sort_columns)
		self._training_indices = training_indices
		self._test_indices = test_indices

	@property
	def training_indices(self):
		return self._training_indices

	@property
	def test_indices(self):
		return self._test_indices

	@property
	def training_data(self):
		"""
		:rtype: DataFrame
		"""
		return self._data.iloc[self._training_indices]

	@property
	def test_data(self):
		"""
		:rtype: DataFrame
		"""
		return self._data.iloc[self._test_indices]

	def _evaluate_estimator(self, estimator, evaluation_type, x_columns=None, y_column=None):
		y_column = y_column or self._y_column
		if y_column is None:
			raise ValueError(f'y_column should be provided')
		x_columns = x_columns or self._x_columns
		if x_columns is None:
			x_columns = [col for col in self.columns if col != y_column]


	def _evaluate(self, estimators, evaluation_type, x_columns=None, y_column=None, echo=1):
		if isinstance(estimators, (list, tuple)):
			estimators = {(i + 1): x for i, x in enumerate(estimators)}

		if isinstance(estimators, dict):
			records = []
			progress_bar = ProgressBar(total=len(estimators), echo=echo)
			progress_amount = 0
			for estimator_id, estimator in estimators.items():
				progress_bar.show(amount=progress_amount, text=f'evaluating estimator {estimator_id}')
				evaluation = self._evaluate_estimator(
					estimator=estimator, evaluation_type=evaluation_type,
					x_columns=x_columns, y_column=y_column
				)
				records.append({
					'estimator_id': estimator_id,
					'estimator_name': estimator.__class__.__name__,
					**evaluation
				})
				progress_amount += 1
			progress_bar.show(amount=progress_amount, text=f'evaluation complete!')
			return DataFrame.from_records(records)

		else:
			return self._evaluate_estimator(
				estimator=estimators, evaluation_type=evaluation_type,
				x_columns=x_columns, y_column=y_column
			)

	def evaluate_regression(self, estimators, x_columns=None, y_column=None, echo=1):
		"""
		evaluates one regressor or more and returns a dictionary or a dataframe of the results
		:param estimators: one or more regressors
		:param x_columns: the x columns (independent variables)
		:param y_column: the y column (dependent variable)
		:param echo: to show progress or not
		:rtype: dict or DataFrame
		"""
		return self._evaluate(
			estimators=estimators, evaluation_type='regression', x_columns=x_columns, y_column=y_column,
			echo=echo
		)

	def evaluate_classification(self, estimators, x_columns=None, y_column=None, echo=1):
		"""
		evaluates one classifier or more and returns a dictionary or a dataframe of the results
		:param estimators: one or more classifiers
		:param x_columns: the x columns (independent variables)
		:param y_column: the y column (dependent variable)
		:param echo: to show progress or not
		:rtype: dict or DataFrame
		"""
		return self._evaluate(
			estimators=estimators, evaluation_type='classification', x_columns=x_columns, y_column=y_column,
			echo=echo
		)

	def display(self, p=None, prefix='', function=None):
		if function is None:
			display = get_display_function()
		else:
			display = function

		if display is False:
			print(f'{prefix}Training:')
			print(self.training_data)
			print(f'{prefix}Test:')
			print(self.test_data)
		else:
			print(f'{prefix}Training:')
			display(self.training_data)
			print(f'{prefix}Test:')
			display(self.test_data)

	def fit(self, model, x_columns=None, y_column=None):
		"""
		:type model: LinearRegression or LogisticRegression
		:type x_columns: list[str]
		:type y_column: str
		"""
		x_columns = x_columns or self._x_columns
		y_column = y_column or self._y_column
		if y_column is None:
			raise MissingArgumentError('y_column should be provided!')
		elif x_columns is None:
			x_columns = [col for col in self._data.columns if col != y_column]

		model.fit(self.training_data[x_columns], self.training_data[y_column])
		return model

	def get_evaluation(self, estimator, estimator_id, x_columns=None, y_column=None, problem_type=None, main_metric=None):
		"""
		:type estimator: LinearRegression or LogisticRegression
		:type estimator_id: int or str or tuple
		:type x_columns: list[str]
		:type y_column: str
		:type problem_type: str or NoneType
		:type main_metric: str or NoneType
		"""
		x_columns = x_columns or self._x_columns
		y_column = y_column or self._y_column
		if y_column is None:
			raise MissingArgumentError('y_column should be provided!')
		elif x_columns is None:
			x_columns = [col for col in self._data.columns if col != y_column]

		evaluation = Evaluation(
			estimator=estimator, estimator_id=estimator_id,
			training_x=self.training_data[x_columns], training_y=self.training_data[y_column],
			test_x=self.test_data[x_columns], test_y=self.test_data[y_column],
			problem_type=problem_type, main_metric=main_metric
		)
		return evaluation





from textwrap import wrap as _wrap
from ._align import align as align_text


def wrap(
		text,
		max_width,
		align='left',
		indentation='',
		prefix='',
		suffix='',
		left_margin=0,
		right_margin=0
):
	remaining_width = max_width - len(indentation) - len(prefix) - len(suffix) - left_margin - right_margin
	inside_width = max_width - len(prefix) - len(suffix)

	return '\n'.join([
		prefix + align_text(
			text=f'{" " * left_margin}{indentation}{line}{" " * right_margin}',
			width=inside_width,
			side=align
		) + suffix
		for line in _wrap(text=text, width=remaining_width)
	])



ENDINGS = ['.', '!', '?', ';']


def count_sentences(text):
	text = text.strip()
	if len(text) == 0:
		return 0

	split_result = None
	for ending in ENDINGS:
		separator = f'{ending} '

		if split_result is None:
			split_result = text.split(separator)

		else:
			split_result = [y for x in split_result for y in x.split(separator)]

	last_is_sentence = text[-1] in ENDINGS

	return len(split_result) - 1 + last_is_sentence



from pandas import DataFrame as PandasDF
from pyspark.sql import DataFrame as SparkDF
from pyspark.sql.types import StringType


def get_string_columns(data):
	"""
	:type data: PandasDF or SparkDF
	:rtype: list[str]
	"""
	if isinstance(data, SparkDF):
		return [
			f.name
			for f in data.schema.fields if isinstance(f.dataType, StringType)
		]
	elif isinstance(data, PandasDF):
		return list(data.select_dtypes(exclude=['bool', 'number', 'datetime64', 'datetime']).columns)

	else:
		raise TypeError(f'data of type {type(data)} is not supported!')



import datetime
from .add_time import add_days
from ..time import get_elapsed


def get_date_mean(dates):
	"""
	:param 	dates: a list or set or tuple of dates
	:type 	dates: list[datetime.date]
	"""

	if len(dates) == 0:
		raise ValueError(f'dates is empty')

	first_date = dates[0]
	total_extra_days = sum([get_elapsed(start=first_date, end=date, unit='day') for date in dates[1:]])
	mean_extra_days = total_extra_days / len(dates)
	return add_days(first_date, days=round(mean_extra_days))
